"""Hierarchical loss for the labels suggested in https://arxiv.org/abs/2210.10929"""

__cmd_doc__ = """Hierarchical loss for the labels"""


import networkx as nx
import numpy as _np
from collections import namedtuple
from functools import partial
from math import log as _log
from pathlib import Path
from typing import Optional, IO, Union, Iterable, Sequence
import dadaptation
from loguru import logger

import torch as _torch
from torch import nn as _nn
import torch.nn.functional as F
from torch.utils.data.dataset import TensorDataset as _TensorDataset
from torch.utils.data import DataLoader as _DataLoader
from torch import Tensor

import vamb.semisupervised_encode as _semisupervised_encode
import vamb.encode as _encode
import vamb.hloss_misc as _hloss
from vamb.taxonomy import ContigTaxonomy


def make_graph(
    taxes: Sequence[Optional[ContigTaxonomy]],
) -> tuple[list[str], dict[str, int], list[int]]:
    G = nx.DiGraph()
    root = "root"
    for contig_taxonomy in taxes:
        if contig_taxonomy is None:
            continue
        if len(contig_taxonomy.ranks) == 0:
            continue
        G.add_edge(root, contig_taxonomy.ranks[0])
        for parent, child in zip(contig_taxonomy.ranks, contig_taxonomy.ranks[1:]):
            G.add_edge(parent, child)
    edges = nx.bfs_edges(G, root)
    # Nodes are unique here, because even though a taxon may appear multiple times
    # in contig taxonomies, its edges will only be added once to the graph
    nodes: list[str] = [root] + [v for _, v in edges]
    ind_nodes = {v: i for i, v in enumerate(nodes)}
    # We have checked during construction of the Taxonomy that node names are unique
    # in that they can't have distinct parents, so this should hold
    assert len(ind_nodes) == len(nodes)
    table_parent: list[int] = []
    for child_node, n in enumerate(nodes):
        if n == root:
            table_parent.append(-1)
        else:
            parent = only(G.predecessors(n))
            parent_index = ind_nodes[parent]
            assert parent_index < ind_nodes[n]
            table_parent.append(parent_index)
    return nodes, ind_nodes, table_parent


def only(itr):
    itr = iter(itr)
    y = next(itr)
    try:
        next(itr)
    except StopIteration:
        return y
    raise ValueError("More than one element in iterator")


def collate_fn_labels_hloss(
    num_categories: int, table_parent, labels
):  # target in form of indices
    return _semisupervised_encode.collate_fn_labels(num_categories, labels)


def collate_fn_concat_hloss(num_categories: int, table_parent, batch):
    a = _torch.stack([i[0] for i in batch])
    b = _torch.stack([i[1] for i in batch])
    c = _torch.stack([i[2] for i in batch])
    d = _torch.stack([i[3] for i in batch])
    e = [i[4] for i in batch]
    return a, b, c, d, collate_fn_labels_hloss(num_categories, table_parent, e)[0]


def collate_fn_semisupervised_hloss(num_categories: int, table_parent, batch):
    a = _torch.stack([i[0] for i in batch])
    b = _torch.stack([i[1] for i in batch])
    c = _torch.stack([i[2] for i in batch])
    d = _torch.stack([i[3] for i in batch])
    e = [i[4] for i in batch]
    f = _torch.stack([i[5] for i in batch])
    g = _torch.stack([i[6] for i in batch])
    h = _torch.stack([i[7] for i in batch])
    m = _torch.stack([i[8] for i in batch])
    n = [i[9] for i in batch]
    return (
        a,
        b,
        c,
        d,
        collate_fn_labels_hloss(num_categories, table_parent, e)[0],
        f,
        g,
        h,
        m,
        collate_fn_labels_hloss(num_categories, table_parent, n)[0],
    )


def make_dataloader_labels_hloss(
    rpkm,
    tnf,
    lengths,
    labels,
    N,
    table_parent,
    batchsize=256,
    destroy=False,
    cuda=False,
):
    _, _, _, _, batchsize, n_workers, cuda = _semisupervised_encode._make_dataset(
        rpkm, tnf, lengths, batchsize=batchsize, destroy=destroy, cuda=cuda
    )
    dataset = _TensorDataset(_torch.Tensor(labels).long())
    dataloader = _DataLoader(
        dataset=dataset,
        batch_size=batchsize,
        drop_last=False,
        shuffle=True,
        num_workers=n_workers,
        pin_memory=cuda,
        collate_fn=partial(collate_fn_labels_hloss, N, table_parent),
    )

    return dataloader


def make_dataloader_concat_hloss(
    rpkm,
    tnf,
    lengths,
    labels,
    N: int,
    table_parent,
    no_filter: bool = True,
    batchsize: int = 256,
    destroy: bool = False,
    cuda: bool = False,
):
    (
        depthstensor,
        tnftensor,
        total_abundance_tensor,
        weightstensor,
        batchsize,
        n_workers,
        cuda,
    ) = _semisupervised_encode._make_dataset(
        rpkm, tnf, lengths, batchsize=batchsize, destroy=destroy, cuda=cuda
    )
    labelstensor = _torch.Tensor(labels).long()
    dataset = _TensorDataset(
        depthstensor, tnftensor, total_abundance_tensor, weightstensor, labelstensor
    )
    dataloader = _DataLoader(
        dataset=dataset,
        batch_size=batchsize,
        drop_last=True,
        shuffle=True,
        num_workers=n_workers,
        pin_memory=cuda,
        collate_fn=partial(collate_fn_concat_hloss, N, table_parent),
    )
    return dataloader


def permute_indices(n_current: int, n_total: int, seed: int):
    rng = _np.random.default_rng(seed)
    x = _np.arange(n_current)
    to_add = n_total // n_current
    to_concatenate = [rng.permutation(x)]
    for _ in range(to_add):
        b = rng.permutation(x)
        to_concatenate.append(b)
    return _np.concatenate(to_concatenate)[:n_total]


def make_dataloader_semisupervised_hloss(
    dataloader_joint,
    dataloader_vamb,
    dataloader_labels,
    N,
    table_parent,
    shapes,
    seed: int,
    batchsize=256,
    cuda=False,
):
    n_total = len(dataloader_vamb.dataset)
    indices_unsup_vamb = permute_indices(len(dataloader_vamb.dataset), n_total, seed)
    indices_unsup_labels = permute_indices(
        len(dataloader_labels.dataset), n_total, seed
    )
    indices_sup = permute_indices(len(dataloader_joint.dataset), n_total, seed)
    dataset_all = _TensorDataset(
        dataloader_vamb.dataset.tensors[0][indices_unsup_vamb],
        dataloader_vamb.dataset.tensors[1][indices_unsup_vamb],
        dataloader_vamb.dataset.tensors[2][indices_unsup_vamb],
        dataloader_vamb.dataset.tensors[3][indices_unsup_vamb],
        dataloader_labels.dataset.tensors[0][indices_unsup_labels],
        dataloader_joint.dataset.tensors[0][indices_sup],
        dataloader_joint.dataset.tensors[1][indices_sup],
        dataloader_joint.dataset.tensors[2][indices_sup],
        dataloader_joint.dataset.tensors[3][indices_sup],
        dataloader_joint.dataset.tensors[4][indices_sup],
    )
    dataloader_all = _DataLoader(
        dataset=dataset_all,
        batch_size=batchsize,
        drop_last=True,
        shuffle=False,
        num_workers=dataloader_joint.num_workers,
        pin_memory=cuda,
        collate_fn=partial(collate_fn_semisupervised_hloss, N, table_parent),
    )
    return dataloader_all


LabelMap = namedtuple("LabelMap", ["to_node", "to_target"])

HierLoss = namedtuple(
    "HierLoss", ["name", "loss_fn", "pred_helper", "pred_fn", "n_labels"]
)

DEFAULT_HIER_LOSS = "flat_softmax"


def init_hier_loss(name, tree):
    CONFIG_LOSSES = dict(
        flat_softmax=HierLoss(
            name="flat_softmax",
            loss_fn=_hloss.FlatSoftmaxNLL(tree),
            pred_helper=_hloss.SumLeafDescendants(tree, strict=False),
            pred_fn=lambda sum_fn, theta: sum_fn(F.softmax(theta, dim=-1), dim=-1),
            n_labels=tree.leaf_mask().nonzero()[0].shape[0],
        ),
        cond_softmax=HierLoss(
            name="cond_softmax",
            loss_fn=_hloss.HierSoftmaxCrossEntropy(tree),
            pred_helper=_hloss.HierLogSoftmax(tree),
            pred_fn=lambda log_softmax_fn, theta: log_softmax_fn(theta).exp(),
            n_labels=tree.num_nodes() - 1,
        ),
        soft_margin=HierLoss(
            name="soft_margin",
            loss_fn=_hloss.MarginLoss(
                tree,
                with_leaf_targets=False,
                hardness="soft",
                margin="incorrect",
                tau=0.01,
            ),
            pred_helper=_hloss.SumDescendants(tree, strict=False),
            pred_fn=lambda sum_fn, theta: sum_fn(F.softmax(theta, dim=-1), dim=-1),
            n_labels=tree.num_nodes(),
        ),
    )
    if name not in CONFIG_LOSSES:
        raise AttributeError(f"Hierarchical loss {name} not found")
    return CONFIG_LOSSES[name]


class VAELabelsHLoss(_semisupervised_encode.VAELabels):
    """Variational autoencoder that encodes only the labels.
    Subclass of one-hot VAELabels class but the labels are structured differently.
    Uses hierarchical loss to utilize the taxonomic tree.
    Instantiate with:
        nlabels: Number of labels

        - The taxonomy graph information:
        nodes: list of node indices in the order of the traversal
        table_parent: a parent for each node in the order of the traversal

        - Hyperparameters:
        nhiddens: List of n_neurons in the hidden layers [None=Auto]
        nlatent: Number of neurons in the latent layer [32]
        alpha: Approximate starting TNF/(CE+TNF) ratio in loss. [None = Auto]
        beta: Multiply KLD by the inverse of this value [200]
        dropout: Probability of dropout on forward pass [0.2]
        cuda: Use CUDA (GPU accelerated training) [False]
    vae.trainmodel(dataloader, nepochs batchsteps, lrate, modelfile)
        Trains the model, returning None
    vae.encode(self, data_loader):
        Encodes the data in the data loader and returns the encoded matrix.
    If alpha or dropout is None and there is only one sample, they are set to
    0.99 and 0.0, respectively
    """

    def __init__(
        self,
        nlabels: int,
        nodes,
        table_parent,
        nhiddens=None,
        nlatent: int = 32,
        alpha: Optional[float] = None,
        beta: float = 200.0,
        dropout: Optional[float] = 0.2,
        hier_loss=DEFAULT_HIER_LOSS,
        cuda: bool = False,
    ):
        super(VAELabelsHLoss, self).__init__(
            nlabels,
            nhiddens=nhiddens,
            nlatent=nlatent,
            alpha=alpha,
            beta=beta,
            dropout=dropout,
            cuda=cuda,
        )
        self.nodes = nodes
        self.table_parent = table_parent
        self.tree = _hloss.Hierarchy(table_parent)
        self.hierloss = init_hier_loss(hier_loss, self.tree)
        self.nlabels = self.hierloss.n_labels
        self.loss_fn = self.hierloss.loss_fn
        self.pred_helper = self.hierloss.pred_helper
        if self.usecuda:
            self.loss_fn = self.loss_fn.cuda()
            self.pred_helper = self.pred_helper.cuda()
        self.pred_fn = partial(self.hierloss.pred_fn, self.pred_helper)
        self.specificity = -self.tree.num_leaf_descendants()
        self.not_trivial = self.tree.num_children() != 1
        self.find_lca = _hloss.FindLCA(self.tree)
        node_subset = _hloss.find_subset_index(self.nodes, self.nodes)
        project_to_subtree = _hloss.find_projection(self.tree, node_subset)
        label_to_node = self.tree.parents()
        label_to_subtree_node = project_to_subtree[label_to_node]
        self.eval_label_map = LabelMap(
            to_node=label_to_subtree_node,
            to_target=label_to_subtree_node,
        )

    def calc_loss(self, labels_in, labels_out, mu, logsigma):
        ce_labels = self.loss_fn(labels_out, labels_in)

        ce_labels_weight = 1.0  # TODO: figure out
        kld = -0.5 * (1 + logsigma - mu.pow(2) - logsigma.exp()).sum(dim=1).mean()
        kld_weight = 1 / (self.nlatent * self.beta)
        loss = ce_labels * ce_labels_weight + kld * kld_weight
        return loss, ce_labels, kld, _torch.tensor(0)

    def trainmodel(
        self,
        dataloader: _DataLoader[tuple[Tensor]],
        nepochs: int = 500,
        lrate: float = 1e-3,
        batchsteps: Optional[list[int]] = [25, 75, 150, 300],
        modelfile: Union[None, str, Path, IO[bytes]] = None,
    ):
        if lrate < 0:
            raise ValueError(f"Learning rate must be positive, not {lrate}")

        if nepochs < 1:
            raise ValueError("Minimum 1 epoch, not {nepochs}")

        if batchsteps is None:
            batchsteps_set: set[int] = set()
        else:
            # First collect to list in order to allow all element types, then check that
            # they are integers
            batchsteps = list(batchsteps)
            if not all(isinstance(i, int) for i in batchsteps):
                raise ValueError("All elements of batchsteps must be integers")
            if max(batchsteps, default=0) >= nepochs:
                raise ValueError("Max batchsteps must not equal or exceed nepochs")
            last_batchsize = dataloader.batch_size * 2 ** len(batchsteps)
            if len(dataloader.dataset) < last_batchsize:  # type: ignore
                raise ValueError(
                    f"Last batch size of {last_batchsize} exceeds dataset length "
                    f"of {len(dataloader.dataset)}. "  # type: ignore
                    "This means you have too few contigs left after filtering to train. "
                    "It is not adviced to run Vamb with fewer than 10,000 sequences "
                    "after filtering. "
                    "Please check the Vamb log file to see where the sequences were "
                    "filtered away, and verify BAM files has sensible content."
                )
            batchsteps_set = set(batchsteps)

        # Get number of features
        # Following line is un-inferrable due to typing problems with DataLoader
        nlabels = dataloader.dataset.tensors[0].shape  # type: ignore
        # optimizer = _Adam(self.parameters(), lr=lrate)
        optimizer = dadaptation.DAdaptAdam(self.parameters(), lr=1, decouple=True)

        logger.info("\tNetwork properties:")
        logger.info(f"\t    CUDA: {self.usecuda}")
        logger.info(f"\t    Alpha: {self.alpha}")
        logger.info(f"\t    Beta: {self.beta}")
        logger.info(f"\t    Dropout: {self.dropout}")
        logger.info(f"\t    N hidden: {', '.join(map(str, self.nhiddens))}")
        logger.info(f"\t    N latent: {self.nlatent}")
        logger.info("\tTraining properties:")
        logger.info(f"\t    N epochs: {nepochs}")
        logger.info(f"\t    Starting batch size: {dataloader.batch_size}")
        batchsteps_string = (
            ", ".join(map(str, sorted(batchsteps_set))) if batchsteps_set else "None"
        )
        logger.info(f"\t    Batchsteps: {batchsteps_string}")
        logger.info(f"\t    Learning rate: {lrate}")
        logger.info(f"\t    N labels: {nlabels}")

        # Train
        for epoch in range(nepochs):
            dataloader = self.trainepoch(
                dataloader, epoch, optimizer, sorted(batchsteps_set)
            )

        # Save weights - Lord forgive me, for I have sinned when catching all exceptions
        if modelfile is not None:
            try:
                self.save(modelfile)
            except:
                pass

        return None


class VAEConcatHLoss(_semisupervised_encode.VAEConcat):
    """Variational autoencoder that encodes the concatenated input of VAMB and labels.
    Subclass of one-hot VAEConcat class but the labels are structured differently.
    Uses hierarchical loss to utilize the taxonomic tree.
    Instantiate with:
        nlabels: Number of labels

        - The taxonomy graph information:
        nodes: list of node indices in the order of the traversal
        table_parent: a parent for each node in the order of the traversal

        - Hyperparameters:
        nhiddens: List of n_neurons in the hidden layers [None=Auto]
        nlatent: Number of neurons in the latent layer [32]
        alpha: Approximate starting TNF/(CE+TNF) ratio in loss. [None = Auto]
        beta: Multiply KLD by the inverse of this value [200]
        dropout: Probability of dropout on forward pass [0.2]
        cuda: Use CUDA (GPU accelerated training) [False]
    vae.trainmodel(dataloader, nepochs batchsteps, lrate, modelfile)
        Trains the model, returning None
    vae.encode(self, data_loader):
        Encodes the data in the data loader and returns the encoded matrix.
    If alpha or dropout is None and there is only one sample, they are set to
    0.99 and 0.0, respectively
    """

    def __init__(
        self,
        nsamples: int,
        nlabels: int,
        nodes,
        table_parent,
        nhiddens: Optional[list[int]] = None,
        nlatent: int = 32,
        alpha=None,
        beta: float = 200.0,
        dropout: Optional[float] = 0.2,
        hier_loss=DEFAULT_HIER_LOSS,
        cuda: bool = False,
    ):
        super(VAEConcatHLoss, self).__init__(
            nsamples,
            nlabels,
            nhiddens=nhiddens,
            nlatent=nlatent,
            alpha=alpha,
            beta=beta,
            dropout=dropout,
            cuda=cuda,
        )
        self.nsamples = nsamples
        self.nodes = nodes
        self.table_parent = table_parent
        self.tree = _hloss.Hierarchy(table_parent)
        self.hierloss = init_hier_loss(hier_loss, self.tree)
        self.nlabels = self.hierloss.n_labels
        self.loss_fn = self.hierloss.loss_fn
        self.pred_helper = self.hierloss.pred_helper
        if self.usecuda:
            self.loss_fn = self.loss_fn.cuda()
            self.pred_helper = self.pred_helper.cuda()
        self.pred_fn = partial(self.hierloss.pred_fn, self.pred_helper)
        self.specificity = -self.tree.num_leaf_descendants()
        self.not_trivial = self.tree.num_children() != 1
        self.find_lca = _hloss.FindLCA(self.tree)
        node_subset = _hloss.find_subset_index(self.nodes, self.nodes)
        project_to_subtree = _hloss.find_projection(self.tree, node_subset)
        label_to_node = self.tree.parents()
        label_to_subtree_node = project_to_subtree[label_to_node]
        self.eval_label_map = LabelMap(
            to_node=label_to_subtree_node,
            to_target=label_to_subtree_node,
        )

    def calc_loss(
        self,
        depths_in,
        depths_out,
        tnf_in,
        tnf_out,
        abundance_in,
        abundance_out,
        labels_in,
        labels_out,
        mu,
        logsigma,
        weights,
    ):
        ab_sse = (abundance_out - abundance_in).pow(2).sum(dim=1)
        # Add 1e-9 to depths_out to avoid numerical instability.
        ce = -((depths_out + 1e-9).log() * depths_in).sum(dim=1)
        sse = (tnf_out - tnf_in).pow(2).sum(dim=1)
        kld = 0.5 * (mu.pow(2)).sum(dim=1)

        # Avoid having the denominator be zero
        if self.nsamples == 1:
            ce_weight = 0.0
        else:
            ce_weight = ((1 - self.alpha) * (self.nsamples - 1)) / (
                self.nsamples * _log(self.nsamples)
            )

        ab_sse_weight = (1 - self.alpha) * (1 / self.nsamples)
        sse_weight = self.alpha / self.ntnf
        kld_weight = 1 / (self.nlatent * self.beta)

        ce_labels = self.loss_fn(labels_out, labels_in)
        ce_labels_weight = 1.0

        loss = (
            ab_sse * ab_sse_weight
            + ce * ce_weight
            + sse * sse_weight
            + ce_labels * ce_labels_weight
            + kld * kld_weight
        ) * weights
        return loss, ce, sse, ce_labels, kld, _torch.tensor(0)


def kld_gauss(p_mu, p_logstd, q_mu, q_logstd):
    loss = (
        q_logstd
        - p_logstd
        + (p_logstd.exp().pow(2) + (p_mu - q_mu).pow(2)) / (2 * q_logstd.exp().pow(2))
        - 0.5
    )
    return loss.mean()


class VAEVAEHLoss(_semisupervised_encode.VAEVAE):
    """Bi-modal variational autoencoder that uses TNFs, abundances and labels.
    It contains three individual encoders (VAMB, labels and concatenated VAMB+labels)
    and two decoders (VAMB and labels).
    Subclass of one-hot VAEVAE class but the labels are structured differently.
    Uses hierarchical loss to utilize the taxonomic tree.
    Instantiate with:
        nlabels: Number of labels

        - The taxonomy graph information:
        nodes: list of node indices in the order of the traversal
        table_parent: a parent for each node in the order of the traversal

        - Hyperparameters:
        nhiddens: List of n_neurons in the hidden layers [None=Auto]
        nlatent: Number of neurons in the latent layer [32]
        alpha: Approximate starting TNF/(CE+TNF) ratio in loss. [None = Auto]
        beta: Multiply KLD by the inverse of this value [200]
        dropout: Probability of dropout on forward pass [0.2]
        cuda: Use CUDA (GPU accelerated training) [False]
    vae.trainmodel(dataloader, nepochs batchsteps, lrate, modelfile)
        Trains the model, returning None
    vae.encode(self, data_loader):
        Encodes the data in the data loader and returns the encoded matrix.
    If alpha or dropout is None and there is only one sample, they are set to
    0.99 and 0.0, respectively
    """

    def __init__(
        self,
        nsamples: int,
        nlabels: int,
        nodes,
        table_parent,
        nhiddens: Optional[list[int]] = None,
        nlatent: int = 32,
        alpha: Optional[float] = None,
        beta: float = 200.0,
        dropout: Optional[float] = 0.2,
        hier_loss=DEFAULT_HIER_LOSS,
        cuda: bool = False,
    ):
        self.usecuda = cuda
        N_l = max(nlabels, 105)
        self.VAEVamb = _encode.VAE(
            nsamples,
            nhiddens=nhiddens,
            nlatent=nlatent,
            alpha=alpha,
            beta=beta,
            dropout=dropout,
            cuda=cuda,
        )
        self.VAELabels = VAELabelsHLoss(
            N_l,
            nodes,
            table_parent,
            nhiddens=nhiddens,
            nlatent=nlatent,
            alpha=alpha,
            beta=beta,
            dropout=dropout,
            hier_loss=hier_loss,
            cuda=cuda,
        )
        self.VAEJoint = VAEConcatHLoss(
            nsamples,
            N_l,
            nodes,
            table_parent,
            nhiddens=nhiddens,
            nlatent=nlatent,
            alpha=alpha,
            beta=beta,
            dropout=dropout,
            hier_loss=hier_loss,
            cuda=cuda,
        )

    @classmethod
    def load(cls, path, nodes, table_parent, cuda=False, evaluate=True):
        """Instantiates a VAE from a model file.
        Inputs:
            path: Path to model file as created by functions VAE.save or
                  VAE.trainmodel.
            cuda: If network should work on GPU [False]
            evaluate: Return network in evaluation mode [True]
        Output: VAE with weights and parameters matching the saved network.
        """

        # Forcably load to CPU even if model was saves as GPU model
        dictionary = _torch.load(
            path, map_location=lambda storage, loc: storage, weights_only=False
        )

        nsamples = dictionary["nsamples"]
        nlabels = dictionary["nlabels"]
        alpha = dictionary["alpha"]
        beta = dictionary["beta"]
        dropout = dictionary["dropout"]
        nhiddens = dictionary["nhiddens"]
        nlatent = dictionary["nlatent"]

        vae = cls(
            nsamples,
            nlabels,
            nodes,
            table_parent,
            nhiddens,
            nlatent,
            alpha,
            beta,
            dropout,
            cuda=cuda,
        )
        vae.VAEVamb.load_state_dict(dictionary["state_VAEVamb"])
        vae.VAELabels.load_state_dict(dictionary["state_VAELabels"])
        vae.VAEJoint.load_state_dict(dictionary["state_VAEJoint"])

        if cuda:
            vae.VAEVamb.cuda()
            vae.VAELabels.cuda()
            vae.VAEJoint.cuda()

        if evaluate:
            vae.VAEVamb.eval()
            vae.VAELabels.eval()
            vae.VAEJoint.eval()

        return vae

    def calc_loss_joint(
        self,
        depths_in,
        depths_out,
        tnf_in,
        tnf_out,
        abundance_in,
        abundance_out,
        labels_in,
        labels_out,
        mu_sup,
        logsigma_sup,
        mu_vamb_unsup,
        logsigma_vamb_unsup,
        mu_labels_unsup,
        logsigma_labels_unsup,
        weights,
    ):
        ab_sse = (abundance_out - abundance_in).pow(2).sum(dim=1)
        # Add 1e-9 to depths_out to avoid numerical instability.
        ce = -((depths_out + 1e-9).log() * depths_in).sum(dim=1)
        sse = (tnf_out - tnf_in).pow(2).sum(dim=1)

        # Avoid having the denominator be zero
        if self.VAEVamb.nsamples == 1:
            ce_weight = 0.0
        else:
            ce_weight = ((1 - self.VAEVamb.alpha) * (self.VAEVamb.nsamples - 1)) / (
                self.VAEVamb.nsamples * _log(self.VAEVamb.nsamples)
            )

        ab_sse_weight = (1 - self.VAEVamb.alpha) * (1 / self.VAEVamb.nsamples)
        sse_weight = self.VAEVamb.alpha / self.VAEVamb.ntnf

        ce_labels = self.VAEJoint.loss_fn(labels_out, labels_in)
        ce_labels_weight = 1.0  # TODO: figure out

        weighed_ab = ab_sse * ab_sse_weight
        weighed_ce = ce * ce_weight
        weighed_sse = sse * sse_weight
        weighed_labels = ce_labels * ce_labels_weight

        reconstruction_loss = weighed_ce + weighed_ab + weighed_sse + weighed_labels

        kld_vamb = kld_gauss(mu_sup, logsigma_sup, mu_vamb_unsup, logsigma_vamb_unsup)
        kld_labels = kld_gauss(
            mu_sup, logsigma_sup, mu_labels_unsup, logsigma_labels_unsup
        )
        kld = kld_vamb + kld_labels
        kld_weight = 1 / (self.VAEVamb.nlatent * self.VAEVamb.beta)
        kld_loss = kld * kld_weight

        loss = (reconstruction_loss + kld_loss) * weights
        return (
            loss.mean(),
            ce.mean(),
            sse.mean(),
            ce_labels.mean(),
            kld_vamb.mean(),
            kld_labels.mean(),
            _torch.tensor(0),
        )


class VAMB2Label(_nn.Module):
    """Label predictor, subclass of torch.nn.Module.

    Instantiate with:
        nsamples: Number of samples in abundance matrix

        - The taxonomy graph information:
        nodes: list of node indices in the order of the traversal
        table_parent: a parent for each node in the order of the traversal

        nhiddens: list of n_neurons in the hidden layers [None=Auto]
        nlatent: Number of neurons in the latent layer [32]
        alpha: Approximate starting TNF/(CE+TNF) ratio in loss. [None = Auto]
        beta: Multiply KLD by the inverse of this value [200]
        dropout: Probability of dropout on forward pass [0.2]
        cuda: Use CUDA (GPU accelerated training) [False]

    vae.trainmodel(dataloader, nepochs batchsteps, lrate, modelfile)
        Trains the model, returning None

    vae.encode(self, data_loader):
        Encodes the data in the data loader and returns the encoded matrix.

    If alpha or dropout is None and there is only one sample, they are set to
    0.99 and 0.0, respectively
    """

    def __init__(
        self,
        nsamples: int,
        nlabels: int,
        nodes,
        table_parent,
        nhiddens: Optional[list[int]] = None,
        alpha: Optional[float] = None,
        beta: float = 200.0,
        dropout: Optional[float] = 0.2,
        hier_loss=DEFAULT_HIER_LOSS,
        cuda: bool = False,
    ):
        if nsamples < 1:
            raise ValueError(f"nsamples must be > 0, not {nsamples}")

        # If only 1 sample, we weigh alpha and nhiddens differently
        if alpha is None:
            alpha = 0.15 if nsamples > 1 else 0.50

        if nhiddens is None:
            nhiddens = [512, 512] if nsamples > 1 else [256, 256]

        if dropout is None:
            dropout = 0.2 if nsamples > 1 else 0.0

        if any(i < 1 for i in nhiddens):
            raise ValueError(f"Minimum 1 neuron per layer, not {min(nhiddens)}")

        if beta <= 0:
            raise ValueError(f"beta must be > 0, not {beta}")

        if not (0 < alpha < 1):
            raise ValueError(f"alpha must be 0 < alpha < 1, not {alpha}")

        if not (0 <= dropout < 1):
            raise ValueError(f"dropout must be 0 <= dropout < 1, not {dropout}")

        super(VAMB2Label, self).__init__()

        # Initialize simple attributes
        self.usecuda = cuda
        self.nsamples = nsamples
        self.ntnf = 103
        self.alpha = alpha
        self.beta = beta
        self.nhiddens = nhiddens
        self.dropout = dropout

        # Initialize lists for holding hidden layers
        self.encoderlayers = _nn.ModuleList()
        self.encodernorms = _nn.ModuleList()
        self.decoderlayers = _nn.ModuleList()
        self.decodernorms = _nn.ModuleList()

        # Add all other hidden layers
        for nin, nout in zip(
            [self.nsamples + self.ntnf + 1] + self.nhiddens, self.nhiddens
        ):
            self.encoderlayers.append(_nn.Linear(nin, nout))
            self.encodernorms.append(_nn.BatchNorm1d(nout))

        self.tree = _hloss.Hierarchy(table_parent)
        self.n_tree_nodes = nlabels

        self.nodes = nodes
        self.table_parent = table_parent
        self.hierloss = init_hier_loss(hier_loss, self.tree)
        self.nlabels = self.hierloss.n_labels
        self.loss_fn = self.hierloss.loss_fn
        self.pred_helper = self.hierloss.pred_helper
        if self.usecuda:
            self.loss_fn = self.loss_fn.cuda()
            self.pred_helper = self.pred_helper.cuda()
        self.pred_fn = partial(self.hierloss.pred_fn, self.pred_helper)
        self.specificity = -self.tree.num_leaf_descendants()
        self.not_trivial = self.tree.num_children() != 1
        self.find_lca = _hloss.FindLCA(self.tree)
        node_subset = _hloss.find_subset_index(self.nodes, self.nodes)
        project_to_subtree = _hloss.find_projection(self.tree, node_subset)
        label_to_node = self.tree.parents()
        label_to_subtree_node = project_to_subtree[label_to_node]
        self.eval_label_map = LabelMap(
            to_node=label_to_subtree_node,
            to_target=label_to_subtree_node,
        )

        # Reconstruction (output) layer
        self.outputlayer = _nn.Linear(self.nhiddens[0], self.nlabels)
        # Activation functions
        self.relu = _nn.LeakyReLU()
        self.dropoutlayer = _nn.Dropout(p=self.dropout)

        if cuda:
            self.cuda()

    def _predict(self, tensor: Tensor) -> tuple[Tensor, Tensor]:
        tensors: list[_torch.Tensor] = list()

        # Hidden layers
        for encoderlayer, encodernorm in zip(self.encoderlayers, self.encodernorms):
            tensor = encodernorm(self.dropoutlayer(self.relu(encoderlayer(tensor))))
            tensors.append(tensor)

        reconstruction = self.outputlayer(tensor)
        labels_out = reconstruction.narrow(1, 0, self.nlabels)
        return labels_out

    def forward(self, depths, tnf, abundances, weights):
        tensor = _torch.cat((depths, tnf, abundances), 1)
        labels_out = self._predict(tensor)
        return labels_out

    def calc_loss(self, labels_in, labels_out):
        ce_labels = self.loss_fn(labels_out, labels_in)
        return ce_labels, _torch.tensor(0)

    def predict(self, data_loader) -> Iterable[tuple[_np.ndarray, _np.ndarray]]:
        self.eval()

        new_data_loader = _encode.set_batchsize(
            data_loader, data_loader.batch_size, encode=True
        )

        with _torch.no_grad():
            for depths, tnf, abundances, weights in new_data_loader:
                # Move input to GPU if requested
                if self.usecuda:
                    depths = depths.cuda()
                    tnf = tnf.cuda()
                    abundances = abundances.cuda()
                    weights = weights.cuda()

                # Evaluate
                labels = self(depths, tnf, abundances, weights)
                with _torch.no_grad():
                    prob = self.pred_fn(labels)
                    if self.usecuda:
                        prob = prob.cpu()
                    pred = _hloss.argmax_with_confidence(
                        self.specificity, prob.numpy(), 0.5, self.not_trivial
                    )
                yield prob.numpy(), pred

    def save(self, filehandle):
        """Saves the VAE to a path or binary opened file. Load with VAE.load

        Input: Path or binary opened filehandle
        Output: None
        """
        state = {
            "nsamples": self.nsamples,
            "alpha": self.alpha,
            "beta": self.beta,
            "dropout": self.dropout,
            "nhiddens": self.nhiddens,
            "nlabels": self.nlabels,
            "state": self.state_dict(),
        }

        _torch.save(state, filehandle)

    @classmethod
    def load(
        cls, path: Union[IO[bytes], str], cuda: bool = False, evaluate: bool = True
    ):
        raise NotImplementedError

    def trainepoch(
        self,
        data_loader,
        epoch: int,
        optimizer,
        batchsteps: list[int],
    ):
        self.train()

        epoch_celoss = 0
        epoch_correct_labels = 0

        if epoch in batchsteps:
            data_loader = _DataLoader(
                dataset=data_loader.dataset,
                batch_size=data_loader.batch_size * 2,
                shuffle=True,
                drop_last=True,
                num_workers=data_loader.num_workers,
                pin_memory=data_loader.pin_memory,
                collate_fn=data_loader.collate_fn,
            )

        for depths_in, tnf_in, abundances_in, weights, labels_in in data_loader:
            depths_in.requires_grad = True
            tnf_in.requires_grad = True
            abundances_in.requires_grad = True
            labels_in.requires_grad = True

            if self.usecuda:
                depths_in = depths_in.cuda()
                tnf_in = tnf_in.cuda()
                abundances_in = abundances_in.cuda()
                weights = weights.cuda()
                labels_in = labels_in.cuda()

            optimizer.zero_grad()

            labels_out = self(depths_in, tnf_in, abundances_in, weights)

            loss, correct_labels = self.calc_loss(labels_in, labels_out)

            loss.mean().backward()
            optimizer.step()

            epoch_celoss += loss.mean().data.item()
            epoch_correct_labels += correct_labels.data.item()

        logger.info(
            f"\tEpoch: {epoch + 1}\tCE: {epoch_celoss / len(data_loader):.7f}\tBatchsize: {data_loader.batch_size}"
        )
        self.eval()
        return data_loader

    def trainmodel(
        self,
        dataloader: _DataLoader[tuple[Tensor, ...]],
        nepochs: int = 500,
        lrate: float = 1e-3,
        batchsteps: Optional[list[int]] = [25, 75, 150, 300],
        modelfile: Union[None, str, Path, IO[bytes]] = None,
    ):
        """Train the autoencoder from depths array and tnf array.

        Inputs:
            dataloader: DataLoader made by make_dataloader
            nepochs: Train for this many epochs before encoding [500]
            lrate: Starting learning rate for the optimizer [0.001]
            batchsteps: None or double batchsize at these epochs [25, 75, 150, 300]
            modelfile: Save models to this file if not None [None]

        Output: None
        """

        if lrate < 0:
            raise ValueError(f"Learning rate must be positive, not {lrate}")

        if nepochs < 1:
            raise ValueError("Minimum 1 epoch, not {nepochs}")

        if batchsteps is None:
            batchsteps_set: set[int] = set()
        else:
            # First collect to list in order to allow all element types, then check that
            # they are integers
            batchsteps = list(batchsteps)
            if not all(isinstance(i, int) for i in batchsteps):
                raise ValueError("All elements of batchsteps must be integers")
            if max(batchsteps, default=0) >= nepochs:
                raise ValueError("Max batchsteps must not equal or exceed nepochs")
            batchsteps_set = set(batchsteps)

        # Get number of features
        # Following line is un-inferrable due to typing problems with DataLoader
        nlabels = dataloader.dataset.tensors[0].shape  # type: ignore
        # optimizer = _Adam(self.parameters(), lr=lrate)
        optimizer = dadaptation.DAdaptAdam(self.parameters(), lr=1, decouple=True)

        logger.info("\tNetwork properties:")
        logger.info(f"\t    CUDA: {self.usecuda}")
        logger.info(f"\t    Hierarchical loss: {self.hierloss.name}")
        logger.info(f"\t    Alpha: {self.alpha}")
        logger.info(f"\t    Beta: {self.beta}")
        logger.info(f"\t    Dropout: {self.dropout}")
        logger.info(f"\t    N hidden: {', '.join(map(str, self.nhiddens))}")
        logger.info("\tTraining properties:")
        logger.info(f"\t    N epochs: {nepochs}")
        logger.info(f"\t    Starting batch size: {dataloader.batch_size}")
        batchsteps_string = (
            ", ".join(map(str, sorted(batchsteps_set))) if batchsteps_set else "None"
        )
        logger.info(f"\t    Batchsteps: {batchsteps_string}")
        logger.info(f"\t    Learning rate: {lrate}")
        logger.info(f"\t    N labels: {nlabels}")

        # Train
        for epoch in range(nepochs):
            dataloader = self.trainepoch(
                dataloader, epoch, optimizer, sorted(batchsteps_set)
            )

        # Save weights - Lord forgive me, for I have sinned when catching all exceptions
        if modelfile is not None:
            try:
                self.save(modelfile)
            except:
                pass

        return None



import sys
import argparse
import vamb
import pathlib

parser = argparse.ArgumentParser(
    description="""Command-line bin creator.
Will read the entire content of the FASTA file into memory - beware.""",
    formatter_class=argparse.RawDescriptionHelpFormatter,
    add_help=False,
)

parser.add_argument("fastapath", help="Path to FASTA file")
parser.add_argument("clusterspath", help="Path to clusters.tsv")
parser.add_argument("minsize", help="Minimum size of bin in bp", type=int, default=0)
parser.add_argument("outdir", help="Directory to create")

if len(sys.argv) == 1:
    parser.print_help()
    sys.exit()

args = parser.parse_args()

# Read in FASTA files only to get its length. This way, we can avoid storing
# in memory contigs for sequences that will never get output anyway
lens: dict[str, int] = dict()
with vamb.vambtools.Reader(args.fastapath) as file:
    for record in vamb.vambtools.byte_iterfasta(file):
        lens[record.identifier] = len(record)

with open(args.clusterspath) as file:
    clusters = vamb.vambtools.read_clusters(file)

clusters = {
    cluster: contigs
    for (cluster, contigs) in clusters.items()
    if sum(lens[c] for c in contigs) >= args.minsize
}

with vamb.vambtools.Reader(args.fastapath) as file:
    vamb.vambtools.write_bins(pathlib.Path(args.outdir), clusters, file, maxbins=None)



from typing import Optional, IO
from pathlib import Path
from vamb.parsecontigs import CompositionMetaData
import numpy as np
from typing import Union

TAXONOMY_HEADER = "contigs\tpredictions"
PREDICTED_TAXONOMY_HEADER = "contigs\tpredictions\tscores"


class ContigTaxonomy:
    """
    Hierarchical taxonomy of some contig.
    If `is_canonical`, the ranks are assumed to be domain, phylum, class,
    order, family, genus, species, in that order.
    The taxonomy may be arbitrarily truncated, e.g. ["Eukaryota", "Chordata"]
    is a valid (canonical) taxonomy for a human.
    """

    __slots__ = ["ranks"]

    def __init__(self, ranks: list[str], is_canonical: bool = False):
        if is_canonical and len(ranks) > 7:
            raise ValueError(
                "For a canonical ContigTaxonomy, there must be at most 7 ranks"
            )

        self.ranks = ranks

    @classmethod
    def from_semicolon_sep(cls, s: str, is_canonical: bool = False):
        return cls(s.split(";"), is_canonical)

    @property
    def genus(self) -> Optional[str]:
        if len(self.ranks) < 6:
            return None
        return self.ranks[5]


class Taxonomy:
    """
    * contig_taxonomies: An Optional[ContigTaxonomy] for every contig given by the
      CompositionMetaData used to instantiate
    * refhash: Refhash of CompositionMetaData used to instantiate
    * is_canonical: If the taxonomy uses the canonical seven ranks
      (domain, phylum, class, order, family, genus, species).
    """

    __slots__ = ["contig_taxonomies", "refhash", "is_canonical"]

    @property
    def nseqs(self) -> int:
        return len(self.contig_taxonomies)

    @classmethod
    def from_file(
        cls, tax_file: Path, metadata: CompositionMetaData, is_canonical: bool
    ):
        observed = cls.parse_tax_file(tax_file, is_canonical)
        return cls.from_observed(observed, metadata, is_canonical)

    @classmethod
    def from_refined_file(
        cls, tax_file: Path, metadata: CompositionMetaData, is_canonical: bool
    ):
        observed = PredictedTaxonomy.parse_tax_file(tax_file, is_canonical)
        observed = [(name, tax.contig_taxonomy) for (name, tax) in observed]
        return cls.from_observed(observed, metadata, is_canonical)

    @classmethod
    def from_observed(
        cls,
        observed_taxonomies: list[tuple[str, ContigTaxonomy]],
        metadata: CompositionMetaData,
        is_canonical: bool,
    ):
        index_of_contigname: dict[str, int] = {
            c: i for (i, c) in enumerate(metadata.identifiers)
        }
        contig_taxonomies: list[Optional[ContigTaxonomy]] = [None] * len(
            metadata.identifiers
        )
        for contigname, taxonomy in observed_taxonomies:
            index = index_of_contigname.get(contigname)
            if index is None:
                raise ValueError(
                    f'When parsing taxonomy, found contigname "{contigname}", '
                    "but no sequence of that name is in the FASTA file"
                )
            existing = contig_taxonomies[index]
            if existing is not None:
                raise ValueError(
                    f'Duplicate contigname when parsing taxonomy: "{contigname}"'
                )
            contig_taxonomies[index] = taxonomy
        return cls(contig_taxonomies, metadata.refhash, is_canonical)

    def __init__(
        self,
        contig_taxonomies: list[Optional[ContigTaxonomy]],
        refhash: bytes,
        is_canonical: bool,
    ):
        self.contig_taxonomies = contig_taxonomies
        self.refhash = refhash
        self.is_canonical = is_canonical
        assert_unambiguous_ranks(self)

    @staticmethod
    def parse_tax_file(
        path: Path, force_canonical: bool
    ) -> list[tuple[str, ContigTaxonomy]]:
        with open(path) as file:
            result: list[tuple[str, ContigTaxonomy]] = []
            header = next(file, None)
            if header is None or not header.startswith(TAXONOMY_HEADER):
                raise ValueError(
                    f"In taxonomy file, expected header to begin with {repr(TAXONOMY_HEADER)}"
                )
            for line in file:
                (contigname, taxonomy, *_) = line.split("\t")
                result.append(
                    (
                        contigname,
                        ContigTaxonomy.from_semicolon_sep(taxonomy, force_canonical),
                    )
                )

        return result


class PredictedContigTaxonomy:
    slots = ["contig_taxonomy", "probs"]

    def __init__(self, tax: ContigTaxonomy, probs: np.ndarray):
        if len(probs) != len(tax.ranks):
            raise ValueError("The length of probs must equal that of ranks")
        # Due to floating point errors, the probabilities may be slightly outside of 0 or 1.
        # We could perhaps validate the values, but that's not likely to be necessary.
        np.clip(probs, a_min=0.0, a_max=1.0, out=probs)
        self.contig_taxonomy = tax
        self.probs = probs


class PredictedTaxonomy:
    "Output of Taxometer"

    __slots__ = ["contig_taxonomies", "refhash", "is_canonical"]

    def __init__(
        self,
        taxonomies: list[PredictedContigTaxonomy],
        metadata: CompositionMetaData,
        is_canonical: bool,
    ):
        if len(taxonomies) != len(metadata.identifiers):
            raise ValueError("Length of taxonomies must match that of identifiers")

        self.contig_taxonomies = taxonomies
        self.refhash = metadata.refhash
        self.is_canonical = is_canonical
        assert_unambiguous_ranks(self)

    def to_taxonomy(self) -> Taxonomy:
        lst: list[Optional[ContigTaxonomy]] = [
            p.contig_taxonomy for p in self.contig_taxonomies
        ]
        return Taxonomy(lst, self.refhash, self.is_canonical)

    @property
    def nseqs(self) -> int:
        return len(self.contig_taxonomies)

    @staticmethod
    def parse_tax_file(
        path: Path, force_canonical: bool
    ) -> list[tuple[str, PredictedContigTaxonomy]]:
        with open(path) as file:
            result: list[tuple[str, PredictedContigTaxonomy]] = []
            lines = filter(None, map(str.rstrip, file))
            header = next(lines, None)
            if header is None or not header.startswith(PREDICTED_TAXONOMY_HEADER):
                raise ValueError(
                    f"In predicted taxonomy file, expected header to begin with {repr(PREDICTED_TAXONOMY_HEADER)}"
                )
            for linenum_minus_two, line in enumerate(lines):
                fields = line.split("\t")
                if len(fields) != 3:
                    raise ValueError(
                        f"Expected 3 fields in line {linenum_minus_two + 2} of file {path}, got {len(fields)}."
                        f"\nLine: {line}"
                    )
                (contigname, taxonomy, scores) = fields
                contig_taxonomy = ContigTaxonomy.from_semicolon_sep(
                    taxonomy, force_canonical
                )
                probs = np.array([float(i) for i in scores.split(";")], dtype=float)
                result.append(
                    (
                        contigname,
                        PredictedContigTaxonomy(contig_taxonomy, probs),
                    )
                )

        return result

    def write_as_tsv(self, file: IO[str], comp_metadata: CompositionMetaData):
        if self.refhash != comp_metadata.refhash:
            raise ValueError(
                "Refhash of comp_metadata and predicted taxonomy must match"
            )
        assert self.nseqs == comp_metadata.nseqs
        print(PREDICTED_TAXONOMY_HEADER, file=file)
        for i in range(self.nseqs):
            tax = self.contig_taxonomies[i]
            ranks_str = ";".join(tax.contig_taxonomy.ranks)
            probs_str = ";".join([str(round(i, 5)) for i in tax.probs])
            print(
                comp_metadata.identifiers[i],
                ranks_str,
                probs_str,
                file=file,
                sep="\t",
            )


def assert_unambiguous_ranks(taxonomy: Union[Taxonomy, PredictedTaxonomy]):
    """
    Ensure that no rank appears at multiple levels in the taxonomy.
    This will mess up some of TaxVamb's algorithms since it's based on the names of
    taxons, and therefore, having a name on two ranks may cause it to be parsed
    as a graph which is not a tree.
    """
    seen_ranks: dict[str, int] = dict()
    parent_of: dict[str, str] = dict()
    for i in taxonomy.contig_taxonomies:
        # May be missing from Taxonomy
        if i is None:
            continue

        if isinstance(i, ContigTaxonomy):
            ranks = i.ranks
        else:
            ranks = i.contig_taxonomy.ranks

        for rank, name in enumerate(ranks):
            if seen_ranks.setdefault(name, rank) != rank:
                raise ValueError(
                    f'Taxonomy is ambiguous: "{name}" appears at multiple ranks'
                )

        for parent, child in zip(ranks, ranks[1:]):
            if parent_of.setdefault(child, parent) != parent:
                raise ValueError(
                    f'Taxonomy is ambiguous: "{child}" has multiple parents'
                )



import vamb
import numpy as np
import os
import itertools

from typing import NewType, Union, Optional
from collections.abc import Sequence, Mapping, Iterable
from pathlib import Path

import argparse
import json

ContigId = NewType("ContigId", int)
BinId = NewType("BinId", int)


def main(
    # Path to output clusters file. Will error if it already exists
    outpath: Path,
    # Path to composition.npz
    composition_path: Path,
    # Path to CheckM2 quality_report.tsv file
    quality_report: dict[str, list],
    # List of paths to clusters.tsv files as output by Vamb.
    # Names of clusters must match those in CheckM2 quality_report,
    # and those of contigs match those in names_npz
    binnings: Sequence[Path],
    # min coverage to consdier 2 bins the same with 100% above this coverage
    min_cov: float,
    # min completentes for bin to be included into the dereplication process
    min_comp: float,
    # max contamination for bin to be included into the dereplication process
    max_cont: float,
    bins_extension: str,
    min_bin_size: int,
) -> None:
    # Load contig names and lengths
    comp = vamb.parsecontigs.Composition.load(composition_path)

    contig_names: list[str] = list(comp.metadata.identifiers)
    assert isinstance(contig_names, list)
    assert isinstance(contig_names[0], str)

    lengths = comp.metadata.lengths
    assert len(lengths) == len(contig_names)
    del comp  # free up memory

    # Load CheckM2
    (bin_names, qualities, bin_by_name) = load_checkm2(
        quality_report, min_comp, max_cont, bins_extension
    )
    # Load bins
    (bin_lengths, union_bins) = load_binnings(
        binnings, contig_names, lengths, bin_by_name, min_bin_size
    )
    del bin_by_name

    dereplicated = dereplicate(union_bins, qualities, lengths, bin_lengths, min_cov)
    del bin_lengths

    if os.path.exists(outpath):
        raise FileExistsError(outpath)

    with open(outpath, "w") as file:
        print(vamb.vambtools.CLUSTERS_HEADER, file=file)
        for bin in dereplicated:
            bin_name = bin_names[bin]
            bin_name = bin_name.replace(".fna", "")
            for contig in union_bins[bin]:
                print(bin_name, contig_names[contig], sep="\t", file=file)


def load_checkm2(
    quality_report: dict[str, list],
    min_completeness: float,
    max_contamination: float,
    bins_extension: str,
) -> tuple[
    list[str],  # Bin names
    list[tuple[float, float]],  # Bin qualities
    dict[str, Optional[BinId]],  # Mapping to binid, if not skipped
]:
    """Extract all bin names and assign them either a BinId, or else None,
    if their completeness/contamination is so bad the bin should be discarded
    """
    # This is None if the bin is to be discarded
    bin_by_name: dict[str, Optional[BinId]] = dict()
    bin_names: list[str] = []
    qualities: list[tuple[float, float]] = []

    # The file looks like this:
    # Name    Completeness    Contamination   Completeness_Model_Used Translation_Table_Used  Additional_Notes
    # AAE_UC_Y_1980340Ccluster_501--AAE_UC_v3_1980340Ccluster_2599    5.18    0.0     Neural Network (Specific Model) 11      None

    for cluster, scores in quality_report.items():
        name = cluster + bins_extension
        comp, cont = scores
        completeness = float(comp) / 100
        contamination = float(cont) / 100
        assert 0.0 <= completeness <= 1.0
        assert 0.0 <= contamination  # can be unbounded

        if completeness >= min_completeness and contamination <= max_contamination:
            bin = BinId(len(bin_names))
            bin_names.append(name)
            qualities.append((completeness, contamination))
            bin_by_name[name] = bin
        else:
            bin_by_name[name] = None

    assert sum(1 for i in bin_by_name.values() if isinstance(i, int)) == len(bin_names)
    return (bin_names, qualities, bin_by_name)


def load_binnings(
    binnings: Sequence[Path],
    contig_names: Sequence[str],
    lengths: np.ndarray,
    bin_by_name: Mapping[str, Optional[BinId]],
    min_bin_size: int,
) -> tuple[list[int], list[set[ContigId]]]:
    """
    Load clusters.tsv files from each binning, and filter away those assigned to be discarded based on CheckM2 data.
    Return bin length and bins, each represented as a set of ContigId
    """
    id_len_of_contig_name: dict[str, tuple[ContigId, int]] = dict()
    for index, (name, length) in enumerate(zip(contig_names, lengths)):
        id_len_of_contig_name[name] = (ContigId(index), length)

    # Load binnings
    n_union_bins = sum(1 for i in bin_by_name.values() if i is not None)

    lengthof = dict(zip(contig_names, lengths))

    union_bins: list[Optional[set[ContigId]]] = [None] * n_union_bins
    for binning_path in binnings:
        with open(binning_path) as file:
            clusters = vamb.vambtools.read_clusters(file)
            clusters_filtered = filterclusters(clusters, lengthof, min_bin_size)
            # filter by clusters larger than 200kbs
            for bin_name, contigs in clusters_filtered.items():
                bin_name += ".fna"
                # None is a valid value, so we use -1 as sentinel for missing
                bin = bin_by_name.get(bin_name, -1)
                if bin == -1:
                    raise ValueError(
                        f"Bin {bin_name} found in binning {binning_path}, but is not scored by CheckM2"
                    )
                # Means: Below threshold, so skip it
                elif bin is None:
                    continue
                else:
                    ids: set[ContigId] = set()
                    for contig in contigs:
                        existing = id_len_of_contig_name.get(contig)
                        if existing is None:
                            raise KeyError(
                                f"Cluster file {binning_path} contain contig {contig}, "
                                "but that name is not present in provided names npz file"
                            )
                        ids.add(existing[0])
                    union_bins[bin] = ids

    bin_lengths: list[int] = []

    for i in union_bins:
        assert isinstance(i, set)
    union_bins_asserted: list[set[ContigId]] = union_bins  # type: ignore

    for contigs in union_bins_asserted:
        bin_lengths.append(sum(lengths[contig] for contig in contigs))

    return (bin_lengths, union_bins_asserted)


def filterclusters(
    clusters: Mapping[str, set], lengthof: Mapping[str, int], min_bin_size: int
) -> Mapping[str, set]:
    filtered_bins = dict()
    for medoid, contigs in clusters.items():
        binsize = sum(lengthof[contig] for contig in contigs)

        if binsize >= min_bin_size:
            filtered_bins[medoid] = contigs

    return filtered_bins


def dereplicate(
    union_bins: Sequence[set[ContigId]],
    qualities: Sequence[tuple[float, float]],
    contig_lengths: np.ndarray,
    bin_lengths: Sequence[int],
    threshold: float,
) -> list[BinId]:
    "Removes bins if they are too similar to another bin. Return list of kept bins"
    assert len(union_bins) == len(qualities) == len(bin_lengths)

    overlapping_pairs = get_overlapping_bin_pairs(get_binsof(union_bins), qualities)
    to_remove = compute_to_remove(
        union_bins, overlapping_pairs, contig_lengths, bin_lengths, threshold
    )
    return [BinId(i) for i in range(len(bin_lengths)) if BinId(i) not in to_remove]


def get_binsof(union_bins: Iterable[Iterable[ContigId]]) -> dict[ContigId, list[BinId]]:
    "Makes a dict from contig -> list of bins the contig is present in, if in multiple bins"
    binsof: dict[ContigId, Union[BinId, list[BinId]]] = dict()
    for bin_int, contigs in enumerate(union_bins):
        bin = BinId(bin_int)
        for contig in contigs:
            existing = binsof.get(contig)
            if existing is None:
                binsof[contig] = bin
            elif isinstance(existing, int):
                binsof[contig] = [existing, bin]
            else:
                assert isinstance(existing, list)
                existing.append(bin)
    return {k: v for (k, v) in binsof.items() if isinstance(v, list)}


def bin_score(completeness: float, contamination: float) -> float:
    return completeness - 5 * contamination


def get_overlapping_bin_pairs(
    binsof: Mapping[ContigId, list[BinId]], qualities: Sequence[tuple[float, float]]
) -> Sequence[tuple[BinId, BinId]]:
    "Get a list of pairs of bins that share at least one contig"
    pairs: set[tuple[BinId, BinId]] = set()
    for overlapping_bins in binsof.values():
        for a, b in itertools.combinations(overlapping_bins, r=2):
            # Order them so we don't have (a, b) and (b, a) as distinct pairs
            if a > b:
                (a, b) = (b, a)
            pairs.add((a, b))

    # Now be sure to order them as (worst, best) depending on score
    # If they tie, then use lexographic order (a, b) we added them
    # in above
    result: list[tuple[BinId, BinId]] = []
    for a, b in pairs:
        score_a = bin_score(*qualities[a])
        score_b = bin_score(*qualities[b])
        if score_a > score_b:
            result.append((b, a))
        else:
            result.append((a, b))

    return result


def compute_to_remove(
    union_bins: Sequence[set[ContigId]],
    overlapping_pairs: Iterable[tuple[BinId, BinId]],
    lengths: np.ndarray,
    bin_lengths: Sequence[int],
    threshold: float,
) -> set[BinId]:
    "Create a list of bins to remove because they overlap with another bin"
    result: set[BinId] = set()
    for bin_a, bin_b in overlapping_pairs:
        if bin_a in result or bin_b in result:
            continue

        intersection = union_bins[bin_a] & union_bins[bin_b]
        int_len = sum(lengths[i] for i in intersection)
        if int_len / min(bin_lengths[bin_a], bin_lengths[bin_b]) >= threshold:
            # We remove an arbitrary one
            result.add(bin_a)
    return result


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--cs_d", type=str, help="path bins_scores dictionary")
    parser.add_argument(
        "--composition", type=Path, help="Path to the composition.npz file"
    )
    parser.add_argument(
        "--output",
        type=str,
        help="Path output clusters generated by dereplicating bins",
    )
    parser.add_argument(
        "--clusters",
        type=str,
        nargs="*",
        help="Path input clusters generated by aamb and vamb",
    )
    parser.add_argument("--cov", type=float, default=0.75, help="Min coverage ")
    parser.add_argument("--comp", type=float, default=0.9, help="Min completeness ")
    parser.add_argument("--cont", type=float, default=0.05, help="Max contamination ")
    parser.add_argument(
        "--bins_extension", type=str, default=".fna", help="Extension of the bins  "
    )
    parser.add_argument(
        "--min_bin_size",
        type=int,
        help="Min bin length to be considered for dereplication ",
    )

    opt = parser.parse_args()
    args = vars(parser.parse_args())
    with open(opt.cs_d) as f:
        cluster_scores = json.load(f)

    main(
        outpath=opt.output,
        composition_path=opt.composition,
        quality_report=cluster_scores,
        binnings=opt.clusters,
        min_cov=opt.cov,
        min_comp=opt.comp,
        max_cont=opt.cont,
        bins_extension=opt.bins_extension,
        min_bin_size=opt.min_bin_size,
    )



__doc__ = """Estimate depths from BAM files of reads mapped to contigs.

Usage:
>>> bampaths = ['/path/to/bam1.bam', '/path/to/bam2.bam', '/path/to/bam3.bam']
>>> rpkms = Abundance.from_file(bampaths, metadata, True, 0.1, 3)
"""

import pycoverm
import os as _os
import numpy as _np
from math import isfinite
from vamb.parsecontigs import CompositionMetaData
from vamb import vambtools
from typing import Optional, TypeVar, Union, IO, Sequence, Iterable
from pathlib import Path
from itertools import zip_longest
import shutil

_ncpu = _os.cpu_count()
DEFAULT_THREADS = 8 if _ncpu is None else _ncpu

A = TypeVar("A", bound="Abundance")


class Abundance:
    "Object representing contig abundance. Contains a matrix and refhash."

    __slots__ = ["matrix", "samplenames", "minid", "refhash"]

    def __init__(
        self,
        matrix: _np.ndarray,
        samplenames: Sequence[str],
        minid: float,
        refhash: bytes,
    ):
        assert matrix.dtype == _np.float32
        assert matrix.ndim == 2
        assert matrix.shape[1] == len(samplenames)
        assert isfinite(minid) and 0.0 <= minid and minid <= 1.0

        self.matrix = matrix
        self.samplenames = _np.array(samplenames, dtype=object)
        self.minid = minid
        self.refhash = refhash

    @property
    def nseqs(self) -> int:
        return len(self.matrix)

    @property
    def nsamples(self) -> int:
        return len(self.samplenames)

    def save(self, io: Union[Path, IO[bytes]]):
        _np.savez_compressed(
            io,
            matrix=self.matrix,
            samplenames=self.samplenames,
            minid=self.minid,
            refhash=self.refhash,
        )

    @classmethod
    def load(
        cls: type[A], io: Union[str, Path, IO[bytes]], refhash: Optional[bytes]
    ) -> A:
        arrs = _np.load(io, allow_pickle=True)
        if "arr_0" in arrs.keys():
            return arrs["arr_0"]  # old format
        abundance = cls(
            vambtools.validate_input_array(arrs["matrix"]),
            arrs["samplenames"],
            arrs["minid"].item(),
            arrs["refhash"].item(),
        )
        if refhash is not None:
            vambtools.RefHasher.verify_refhash(
                abundance.refhash,
                refhash,
                "the loaded Abundance object",
                "the given refhash",
                None,
            )

        return abundance

    @classmethod
    def from_files(
        cls: type[A],
        paths: list[Path],
        cache_directory: Optional[Path],
        comp_metadata: CompositionMetaData,
        verify_refhash: bool,
        minid: float,
        nthreads: int,
    ) -> A:
        """Input:
        paths: List of paths to BAM files
        cache_directory: Where to store temp parts of the larger matrix, if reading multiple
           BAM files in chunks. Required if len(paths) > min(16, nthreads)
        comp_metadata: CompositionMetaData of sequence catalogue used to make BAM files
        verify_refhash: Whether to verify composition and BAM references are the same
        minid: Discard any reads with nucleotide identity less than this
        nthreads: Use this number of threads for coverage estimation
        """
        if minid < 0 or minid > 1:
            raise ValueError(f"minid must be between 0 and 1, not {minid}")

        # Workaround: Currently pycoverm has a bug where it filters contigs when mindid == 0
        # (issue #7). Can be solved by setting it to a low value
        minid = minid if minid > 0.001 else 0.001

        if nthreads < 1:
            raise ValueError(f"nthreads must be > 0, not {nthreads}")

        chunksize = min(nthreads, len(paths))

        # We cap it to 16 threads, max. This will prevent pycoverm from consuming a huge amount
        # of memory if given a crapload of threads, and most programs will probably be IO bound
        # when reading 16 files at a time.
        chunksize = min(chunksize, 16)

        # If it can be done in memory, do so
        if chunksize >= len(paths):
            (matrix, refhash) = cls.run_pycoverm(
                paths,
                minid,
                comp_metadata.refhash if verify_refhash else None,
                comp_metadata.identifiers if verify_refhash else None,
                comp_metadata.mask,
            )
            vambtools.mask_lower_bits(matrix, 12)
            return cls(matrix, [str(p) for p in paths], minid, refhash)
        # Else, we load it in chunks, then assemble afterwards
        else:
            if cache_directory is None:
                raise ValueError(
                    "If min(16, nthreads) < len(paths), cache_directory must not be None"
                )
            return cls.chunkwise_loading(
                paths,
                cache_directory,
                chunksize,
                minid,
                comp_metadata.refhash if verify_refhash else None,
                comp_metadata.identifiers if verify_refhash else None,
                comp_metadata.mask,
            )

    @classmethod
    def chunkwise_loading(
        cls: type[A],
        paths: list[Path],
        cache_directory: Path,
        nthreads: int,
        minid: float,
        target_refhash: Optional[bytes],
        target_identifiers: Optional[Iterable[str]],
        mask: _np.ndarray,
    ) -> A:
        _os.makedirs(cache_directory)

        chunks = [
            (i, min(len(paths), i + nthreads)) for i in range(0, len(paths), nthreads)
        ]
        filenames = [
            _os.path.join(cache_directory, str(i) + ".npz") for i in range(len(chunks))
        ]
        assert len(chunks) > 1

        # Load from BAM and store them chunkwise
        refhash = None
        for filename, (chunkstart, chunkstop) in zip(filenames, chunks):
            (matrix, refhash) = cls.run_pycoverm(
                paths[chunkstart:chunkstop],
                minid,
                target_refhash,
                target_identifiers,
                mask,
            )
            vambtools.write_npz(filename, matrix)

        # Initialize matrix, the load them chunkwise. Delete the temp files when done
        matrix = _np.empty((mask.sum(), len(paths)), dtype=_np.float32)
        for filename, (chunkstart, chunkstop) in zip(filenames, chunks):
            matrix[:, chunkstart:chunkstop] = vambtools.read_npz(filename)
        vambtools.mask_lower_bits(matrix, 12)

        shutil.rmtree(cache_directory)

        assert refhash is not None
        return cls(matrix, [str(p) for p in paths], minid, refhash)

    @staticmethod
    def run_pycoverm(
        paths: list[Path],
        minid: float,
        target_refhash: Optional[bytes],
        target_identifiers: Optional[Iterable[str]],
        mask: _np.ndarray,
    ) -> tuple[_np.ndarray, bytes]:
        (headers, coverage) = pycoverm.get_coverages_from_bam(
            [str(p) for p in paths],
            threads=len(paths),
            min_identity=minid,
            # Note: pycoverm's trim_upper=0.1 is same as CoverM trim-upper 90.
            trim_upper=0.1,
            trim_lower=0.1,
        )

        assert coverage.shape == (len(headers), len(paths))

        # Filter length, using comp_metadata's mask, which has been set by minlength
        if len(mask) != len(headers):
            raise ValueError(
                f"CompositionMetaData used to create Abundance object was created with {len(mask)} sequences, "
                f"but number of reference sequences in BAM files are {len(headers)}. "
                "Make sure the BAM files were created by mapping to the same FASTA file "
                "which you used to create the Composition object."
            )

        headers = [h for (h, m) in zip(headers, mask) if m]
        vambtools.numpy_inplace_maskarray(coverage, mask)
        refhash = vambtools.RefHasher.hash_refnames(headers)

        if target_identifiers is None:
            identifier_pairs = None
        else:
            identifier_pairs = (headers, target_identifiers)

        if target_refhash is not None:
            vambtools.RefHasher.verify_refhash(
                refhash, target_refhash, "FASTA file", "BAM", identifier_pairs
            )

        return (coverage, refhash)

    @classmethod
    def from_tsv(cls: type[A], path: Path, comp_metadata: CompositionMetaData) -> A:
        seen_identifiers: list[str] = []
        with open(path) as file:
            try:
                header = next(file)
            except StopIteration:
                err = ValueError(f'Empty abundance TSV path at "{path}"')
                raise err from None
            columns = header.splitlines()[0].split("\t")
            if len(columns) < 1:
                raise ValueError(
                    f'Expected at least 1 column in abundance TSV file at "{path}"'
                )
            if columns[0] != "contigname":
                raise ValueError('First column in header must be "contigname"')
            samples = columns[1:]
            n_samples = len(samples)
            matrix = _np.empty((comp_metadata.nseqs, n_samples), dtype=_np.float32)
            matrix_row = 0
            found_empty = False
            for line_number_minus_two, (line, should_keep) in enumerate(
                zip_longest(file, comp_metadata.mask)
            ):
                if line is not None:
                    if not line.strip():
                        found_empty = True
                    elif found_empty:
                        raise ValueError(
                            "Found an empty line not at end of abundance TSV file"
                            f'"{path}"'
                        )
                else:
                    raise ValueError(
                        f'Too few rows in abundance TSV file "{path}", expected '
                        f"{len(comp_metadata.mask) + 1}, got {line_number_minus_two + 1}"
                    )
                if should_keep is None:
                    raise ValueError(
                        f'Too many rows in abundance TSV file "{path}", expected '
                        f"{len(comp_metadata.mask) + 1} sequences, got at least "
                        f"{line_number_minus_two + 2}"
                    )

                stripped = line.rstrip()
                if not should_keep:
                    continue
                fields = stripped.split("\t")
                if len(fields) != n_samples + 1:
                    raise ValueError(
                        f'In abundance TSV file "{path}", on line {line_number_minus_two + 2}'
                        f", expected {n_samples + 1} columns, found {len(fields)}"
                    )
                for i in range(n_samples):
                    matrix[matrix_row, i] = float(fields[i + 1])
                matrix_row += 1
                seen_identifiers.append(fields[0])

        vambtools.RefHasher.verify_refhash(
            vambtools.RefHasher.hash_refnames(seen_identifiers),
            comp_metadata.refhash,
            "abundance TSV",
            "composition",
            (seen_identifiers, comp_metadata.identifiers),
        )

        return cls(matrix, samples, 0.0, comp_metadata.refhash)



import numpy as np
import os
import json
import argparse

from typing import cast


def get_cluster_score_bin_path(
    path_checkm_all: str, path_bins: str, bins: set[str]
) -> tuple[dict[str, tuple[float, float]], dict[str, str]]:
    """Given CheckM has been run for all samples, create 2 dictionaries:
    - {bin:path_bin}
    - {bin:[completeness, contamination]}"""
    cluster_score: dict[str, tuple[float, float]] = dict()
    bin_path: dict[str, str] = dict()
    for sample in os.listdir(path_checkm_all):
        path_quality_s = os.path.join(path_checkm_all, sample, "quality_report.tsv")
        c_com_con = np.loadtxt(
            path_quality_s,
            delimiter="\t",
            skiprows=1,
            usecols=(0, 1, 2),
            dtype=str,
            ndmin=2,
        )

        for row in c_com_con:
            cluster, com, con = row
            cluster = cast(str, cluster)
            com, con = float(com), float(con)
            bin_name = cluster + ".fna"
            if bin_name in bins:
                cluster_score[cluster] = (com, con)
                bin_path[cluster + ".fna"] = os.path.join(
                    path_bins, sample, cluster + ".fna"
                )
    return cluster_score, bin_path


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--s", type=str, help="path checkm2 that contains all samples")
    parser.add_argument("--b", type=str, help="path all bins ")
    parser.add_argument(
        "--cs_d", type=str, help="cluster_score dictionary will be stored here"
    )
    parser.add_argument(
        "--bp_d", type=str, help="bin_path dictionary will be stored here "
    )

    opt = parser.parse_args()

    bins_set = set()
    for sample in os.listdir(opt.b):
        for bin_ in os.listdir(os.path.join(opt.b, sample)):
            if ".fna" in bin_:
                bins_set.add(bin_)

    cluster_score, bin_path = get_cluster_score_bin_path(opt.s, opt.b, bins_set)
    with open(opt.cs_d, "w") as f:
        json.dump(cluster_score, f)

    with open(opt.bp_d, "w") as f:
        json.dump(bin_path, f)



import numpy as np
import json
import argparse


def update_cluster_score_bin_path(
    path_checkm_ripped: str, cluster_score: dict[str, tuple[float, float]]
):
    c_com_con = np.loadtxt(
        path_checkm_ripped,
        delimiter="\t",
        skiprows=1,
        usecols=(0, 1, 2),
        dtype=str,
        ndmin=2,
    )
    for row in c_com_con:
        cluster, com, con = row
        if "--" in cluster:
            continue
        com, con = float(com), float(con)
        print(cluster, "scores were", cluster_score[cluster])

        cluster_score[cluster] = (com, con)
        print("and now are", cluster_score[cluster])
    return cluster_score


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--s",
        type=str,
        help="path checkm2 that contains quality_report.tsv file for ripped bins",
    )
    parser.add_argument(
        "--cs_d",
        type=str,
        help="cluster_score dictionary path  ",
    )
    parser.add_argument(
        "--cs_d_o",
        type=str,
        help="cluster_score dictionary path updated, updated with tthe information for clusters that where ripped either becuase of meaningless edges or when making the component lenght <= 2  ",
    )

    opt = parser.parse_args()

    with open(opt.cs_d) as f:
        cluster_score = json.load(f)

    cluster_score_ = update_cluster_score_bin_path(opt.s, cluster_score)

    with open(opt.cs_d_o, "w") as f:
        json.dump(cluster_score_, f)



import unittest
import numpy as np
from hashlib import md5

import vamb


class TestClusterer(unittest.TestCase):
    # This seed has been set just so the unit tests runs faster.
    # How many iterations of the clustering depends on the input data
    rng = np.random.RandomState(5)
    data = rng.random((1024, 40)).astype(np.float32)
    lens = rng.randint(500, 1000, size=1024)

    def test_bad_params(self):
        with self.assertRaises(ValueError):
            vamb.cluster.ClusterGenerator(self.data.astype(np.float64), self.lens)

        with self.assertRaises(ValueError):
            vamb.cluster.ClusterGenerator(self.data, self.lens, maxsteps=0)

        with self.assertRaises(ValueError):
            vamb.cluster.ClusterGenerator(self.data, self.lens, windowsize=0)

        with self.assertRaises(ValueError):
            vamb.cluster.ClusterGenerator(self.data, self.lens, minsuccesses=0)

        with self.assertRaises(ValueError):
            vamb.cluster.ClusterGenerator(
                self.data, self.lens, minsuccesses=5, windowsize=4
            )

        with self.assertRaises(ValueError):
            vamb.cluster.ClusterGenerator(
                np.random.random((0, 40)), np.array([], dtype=int)
            )

    def test_basics(self):
        clstr = vamb.cluster.ClusterGenerator(self.data, self.lens)
        self.assertIs(clstr, iter(clstr))

        x = next(clstr)
        self.assertIsInstance(x, vamb.cluster.Cluster)

        clusters = list(clstr)
        clusters.append(x)

        # All members are clustered
        self.assertEqual(sum(map(lambda x: len(x.members), clusters)), len(self.data))

        # Elements of members are exactly the matrix row indices
        mems = set()
        for i in clusters:
            mems.update(i.members)
        self.assertEqual(mems, set(range(len(self.data))))

    def test_detruction(self):
        copy = self.data.copy()
        clstr = vamb.cluster.ClusterGenerator(self.data, self.lens)
        self.assertTrue(np.any(np.abs(self.data - clstr.matrix.numpy()) > 0.001))
        clstr = vamb.cluster.ClusterGenerator(copy, self.lens, destroy=True)
        self.assertTrue(np.all(np.abs(copy - clstr.matrix.numpy()) < 1e-6))
        self.assertTrue(np.any(np.abs(self.data - clstr.matrix.numpy()) > 0.001))

    @staticmethod
    def xor_rows_hash(matrix):
        m = np.frombuffer(matrix.copy().data, dtype=np.uint32)
        m.shape = matrix.shape
        v = m[0]
        for i in range(1, len(m)):
            v ^= m[i]
        return md5(v).digest().hex()

    def test_normalization(self):
        hash_before = md5(self.data.data.tobytes()).digest().hex()
        vamb.cluster.ClusterGenerator(self.data, self.lens)
        self.assertEqual(hash_before, md5(self.data.data.tobytes()).digest().hex())
        cp = self.data.copy()
        vamb.cluster.ClusterGenerator(cp, self.lens, destroy=True)
        hash_after = md5(cp.data.tobytes()).digest().hex()
        self.assertNotEqual(hash_before, hash_after)

        # Rows are permuted by the clusterer. We use xor to check the rows
        # are still essentially the same.
        before_xor = self.xor_rows_hash(cp)
        vamb.cluster.ClusterGenerator(cp, self.lens, destroy=True, normalized=True)
        self.assertEqual(before_xor, self.xor_rows_hash(cp))

    def test_cluster(self):
        x = next(vamb.cluster.ClusterGenerator(self.data, self.lens))
        self.assertIsInstance(x.members, np.ndarray)



import numpy as np
import argparse
import vamb
from pathlib import Path


def write_abundances(
    mask_refhash: Path, bampath: Path, min_identity: float, outfile: Path
):
    """For every sample, compute the abundances given the mask and refhashes"""
    loadnpz = np.load(mask_refhash)
    refhash = loadnpz["refhash"]
    mask = loadnpz["mask"]
    refhash = refhash.reshape(1)[0]
    (abundance, _) = vamb.parsebam.Abundance.run_pycoverm(
        paths=[bampath],
        minid=min_identity,
        target_refhash=refhash,
        target_identifiers=None,
        mask=mask,
    )
    vamb.vambtools.write_npz(outfile, abundance.ravel())


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--msk", type=Path, help="mask refhash")
    parser.add_argument("--b", type=Path, help=" bam path")
    parser.add_argument("--min_id", type=float, help="min identity for alignment")
    parser.add_argument("--out", type=Path, help="abundances outfile")

    opt = parser.parse_args()

    write_abundances(opt.msk, opt.b, opt.min_id, opt.out)



# Overview
# We use pyrodigal to predict genes in every contigs not filtered away by
# the given mask, then use pyhmmer to predict single copy marker genes (SCGs)
# on the genes, hence getting a contig => list[SCG] mapping.
# Pyrodigal/pyhmmer is a bottleneck, so we run in parallel processes.
# To avoid inter-process communication overhead, we first split the input
# FASTA files to N files, then we have each process work on the files independently.

from vamb.vambtools import FastaEntry, Reader, RefHasher, byte_iterfasta
import pyrodigal
import pyhmmer
from multiprocessing.pool import Pool
import os
import itertools
from pathlib import Path
from typing import NewType, Sequence, Union, IO, Optional, Iterable
import shutil
from collections import defaultdict
import json
import numpy as np
from loguru import logger

MarkerID = NewType("MarkerID", int)
MarkerName = NewType("MarkerName", str)
ContigID = NewType("ContigID", int)
ContigName = NewType("ContigName", str)


class Markers:
    """
    The set of marker genes predicted for a collection of contigs.
    Instantiate using `Markers.from_files`, or load using `Markers.load`.
    Like Abundance objects, Markers carry a refhash to check that the markers correspond
    to the same sequences used to create the markers.
    Access the markers with `markers.markers`, a `list[Optional[np.array]]`, with one
    element for each contig. The element is `None` if there are no markers, else a list
    of marker genes present in the contig.
    The marker genes are stored as integers - the name of a marker `i` can be gotten using
    `markers.marker_names[i]`.
    In each contig, markers are deduplicated, so at most 1 of each marker is found
    in each contig.
    """

    __slots__ = ["markers", "marker_names", "refhash"]

    def __init__(
        self,
        markers: list[Optional[np.ndarray]],
        # Some IDs map to multiple names, if they act as the same SCG in the cell
        marker_names: list[list[MarkerName]],
        refhash: bytes,
    ):
        if len(set(itertools.chain.from_iterable(marker_names))) != sum(
            len(i) for i in marker_names
        ):
            raise ValueError("Marker names are not unique, but must be")

        self.markers = markers
        self.marker_names = marker_names
        self.refhash = refhash

    @property
    def n_markers(self):
        return len(self.marker_names)

    @property
    def n_seqs(self):
        return len(self.markers)

    def score_bin(self, indices: Iterable[int]) -> tuple[float, float]:
        counts = np.zeros(self.n_markers, dtype=np.uint8)
        for i in indices:
            mkrs = self.markers[i]
            if mkrs is None:
                continue
            for m in mkrs:
                counts[m] += 1

        n_unique = (counts > 0).sum()
        completeness = n_unique / self.n_markers
        contamination = (counts.sum() - n_unique) / self.n_markers
        return (completeness, contamination)

    def save(self, io: Union[Path, str, IO[str]]):
        representation = {
            "markers": [i if i is None else i.tolist() for i in self.markers],
            "marker_names": self.marker_names,
            "refhash": self.refhash.hex(),
        }
        # Check we didn't forget any fields
        assert len(representation) == len(self.__slots__)
        if isinstance(io, Path) or isinstance(io, str):
            with open(io, "w") as file:
                json.dump(representation, file)

        else:
            json.dump(representation, io)

    @classmethod
    def load(cls, io: Union[Path, str, IO[str]], refhash: Optional[bytes]):
        if isinstance(io, Path) or isinstance(io, str):
            with open(io, "rb") as file:
                representation = json.load(file)
        else:
            representation = json.load(io)
        observed_refhash = bytes.fromhex(representation["refhash"])
        if refhash is not None:
            RefHasher.verify_refhash(
                refhash=observed_refhash,
                target_refhash=refhash,
                observed_name="Loaded markers",
                target_name=None,
                identifiers=None,
            )
        markers_as_arrays = [
            i if i is None else np.array(i, dtype=np.uint8)
            for i in representation["markers"]
        ]

        return cls(markers_as_arrays, representation["marker_names"], observed_refhash)

    @classmethod
    def from_files(
        cls,
        contigs: Path,
        hmm_path: Path,
        contignames: Sequence[str],
        tmpdir_to_create: Path,
        n_processes: int,
        target_refhash: Optional[bytes],
    ):
        """
        Create the Markers from input files:
        `contigs`: Path to a FASTA file with all contigs, gzipped or not.
        `hmm_path`: Path to a HMMER .hmm file with the markers. Note: Currently,
          this file can contain at most 256 markers, though this restriction can
          be lifted if necessary

        The `fasta_entry_mask` is a boolean mask of which contigs in the FASTA
        file to include. This affects the refhash which is only computed for
        the contigs not filtered away.
        If the target refhash is not None, and the computed reference hash does not
        match, an exception is thrown. See vamb.vambtools.RefHasher.
        """
        n_processes = cap_processes(n_processes)
        with open(hmm_path, "rb") as file:
            hmms = list(pyhmmer.plan7.HMMFile(file))
        (_, marker_names) = get_name_to_id(hmms)

        (refhash, paths) = split_file(
            contigs,
            contignames,
            tmpdir_to_create,
            n_processes,
        )

        if target_refhash is not None:
            RefHasher.verify_refhash(
                refhash, target_refhash, "Markers FASTA file", None, None
            )

        index_of_name = {
            ContigName(n): ContigID(i) for (i, n) in enumerate(contignames)
        }
        marker_list: list[Optional[np.ndarray]] = [None] * len(contignames)
        with Pool(n_processes) as pool:
            for sub_result in pool.imap_unordered(
                work_per_process,
                list(zip(paths, itertools.repeat(hmms))),
            ):
                for contig_name, markers in sub_result:
                    marker_list[index_of_name[contig_name]] = markers

        shutil.rmtree(tmpdir_to_create)
        markers = cls(marker_list, marker_names, refhash)

        return markers


def cap_processes(processes: int) -> int:
    if processes < 1:
        raise ValueError(f"Must use at least 1 process, not {processes}")
    # Cap processes, because most OSs cap the number of open file handles,
    # and we need one file per process when splitting FASTA file
    elif processes > 64:
        logger.warning(f"Processes set to {processes}, capping to 64")
        return 64
    return processes


# Some markers have different names, but should be treated as the same SCG.
NORMALIZE_MARKER_TRANS_DICT = {
    "TIGR00388": "TIGR00389",
    "TIGR00471": "TIGR00472",
    "TIGR00408": "TIGR00409",
    "TIGR02386": "TIGR02387",
}


def split_file(
    input: Path,
    contignames: Sequence[str],
    tmpdir_to_create: Path,
    n_splits: int,
) -> tuple[bytes, list[Path]]:
    names = set(contignames)
    os.mkdir(tmpdir_to_create)
    paths = [tmpdir_to_create.joinpath(str(i)) for i in range(n_splits)]
    filehandles = [open(path, "w") for path in paths]
    refhasher = RefHasher()
    with Reader(input) as infile:
        for i, (outfile, record) in enumerate(
            zip(
                itertools.cycle(filehandles),
                filter(lambda x: x.identifier in names, byte_iterfasta(infile)),
            )
        ):
            refhasher.add_refname(record.identifier)
            print(record.format(), file=outfile)

    for filehandle in filehandles:
        filehandle.close()
    refhash = refhasher.digest()
    return (refhash, paths)


def process_chunk(
    chunk: list[FastaEntry],
    hmms: list[pyhmmer.plan7.HMM],
    name_to_id: dict[MarkerName, MarkerID],
    finder: pyrodigal.GeneFinder,
) -> list[tuple[ContigName, np.ndarray]]:
    # We temporarily store them as sets in order to deduplicate. While single contigs
    # may have duplicate markers, it makes no sense to count this as contamination,
    # because we are not about to second-guess the assembler's job of avoiding
    # chimeric sequences.
    markers: defaultdict[ContigName, set[MarkerID]] = defaultdict(set)
    alphabet = pyhmmer.easel.Alphabet.amino()
    digitized: list[pyhmmer.easel.DigitalSequence] = []
    for record in chunk:
        for gene in finder.find_genes(record.sequence):
            seq = pyhmmer.easel.TextSequence(
                name=record.identifier.encode(), sequence=gene.translate()
            ).digitize(alphabet)
            digitized.append(seq)

    for hmm, top_hits in zip(hmms, pyhmmer.hmmsearch(hmms, digitized)):
        marker_name = MarkerName(hmm.name.decode())
        marker_id = name_to_id[marker_name]
        # We need this score cutoff, which is stored in the HMM file to remove the large
        # number of false positives from HMMER
        score_cutoff = hmm.cutoffs.trusted1
        assert score_cutoff is not None
        for hit in top_hits:
            if hit.score >= score_cutoff:
                markers[ContigName(hit.name.decode())].add(marker_id)

    return [
        (name, np.array(list(ids), dtype=np.uint8)) for (name, ids) in markers.items()
    ]


def work_per_process(
    args: tuple[Path, list[pyhmmer.plan7.HMM]],
) -> list[tuple[ContigName, np.ndarray]]:
    (contig_path, hmms) = args

    (name_to_id, _) = get_name_to_id(hmms)

    # Chunk up the FASTA file for memory efficiency reasons, while still
    # allowing pyhmmer to scan multiple sequences at once for speed
    chunk: list[FastaEntry] = []
    result: list[tuple[ContigName, np.ndarray]] = []
    finder = pyrodigal.GeneFinder(meta=True)
    with open(contig_path, "rb") as file:
        for record in byte_iterfasta(file):
            chunk.append(record)
            if len(chunk) == 2048:
                result.extend(process_chunk(chunk, hmms, name_to_id, finder))
                chunk.clear()
        result.extend(process_chunk(chunk, hmms, name_to_id, finder))

    return result


def get_name_to_id(
    hmms: list[pyhmmer.plan7.HMM],
) -> tuple[dict[MarkerName, MarkerID], list[list[MarkerName]]]:
    name_to_id: dict[MarkerName, MarkerID] = dict()
    for hmm in hmms:
        name = hmm.name.decode()
        if name in NORMALIZE_MARKER_TRANS_DICT:
            continue
        name_to_id[MarkerName(name)] = MarkerID(len(name_to_id))
    for old_name, new_name in NORMALIZE_MARKER_TRANS_DICT.items():
        name_to_id[MarkerName(old_name)] = name_to_id[MarkerName(new_name)]

    if len(set(name_to_id.values())) > 256:
        raise ValueError("Maximum 256 marker IDs")

    id_to_names: defaultdict[MarkerID, list[MarkerName]] = defaultdict(list)
    for n, i in name_to_id.items():
        id_to_names[i].append(n)
    marker_names = [id_to_names[MarkerID(i)] for i in range(len(id_to_names))]

    return name_to_id, marker_names



import io
import unittest
import random
import numpy as np

import testtools
from vamb.parsecontigs import Composition, CompositionMetaData


class TestReadContigs(unittest.TestCase):
    records = []
    large_io = io.BytesIO()
    io = io.BytesIO()

    @classmethod
    def setUpClass(cls):
        rng = random.Random()
        for i in range(random.randrange(1400, 1500)):
            cls.records.append(testtools.make_randseq(rng, 400, 600))

        for i in cls.records:
            cls.io.write(i.format().encode())
            cls.io.write(b"\n")

        for i in range(25_000):
            record = testtools.make_randseq(rng, 250, 300)
            cls.large_io.write(record.format().encode())
            cls.large_io.write(b"\n")

    def setUp(self):
        self.io.seek(0)
        self.large_io.seek(0)

    def test_only_ns(self):
        file = io.BytesIO()
        file.write(b">abc\n")
        file.write(b"N" * 2500)
        file.write(b"\n")
        file.seek(0)

        with self.assertRaises(ValueError):
            Composition.from_file(file)

    def test_unique_names(self):
        with self.assertRaises(ValueError):
            CompositionMetaData(
                np.array(["foo", "foo"], dtype=object),
                np.array([1000, 1000]),
                np.array([True, True], dtype=bool),
                1000,
            )

    def test_filter_minlength(self):
        minlen = 500
        composition = Composition.from_file(self.io, minlength=450)
        md = composition.metadata
        hash1 = md.refhash

        composition.filter_min_length(minlen)
        n_initial_seq = md.nseqs

        hash2 = md.refhash
        self.assertNotEqual(hash1, hash2)
        self.assertEqual(len(md.identifiers), len(md.lengths))
        self.assertEqual(md.nseqs, md.mask.sum())
        self.assertLessEqual(minlen, composition.metadata.lengths.min(initial=minlen))
        self.assertEqual(len(md.mask), len(self.records))

        # NB: Here we filter metadata without filtering the composition.
        # That means from this point on, the metadata and comp is out of sync,
        # and comp is invalid.
        md.filter_min_length(minlen + 50)
        self.assertEqual(len(md.identifiers), len(md.lengths))
        self.assertEqual(md.nseqs, md.mask.sum())
        self.assertLessEqual(
            minlen, composition.metadata.lengths.min(initial=minlen + 50)
        )
        self.assertEqual(len(md.mask), len(self.records))
        self.assertLess(md.nseqs, n_initial_seq)

        hash3 = md.refhash
        md.filter_min_length(minlen - 50)
        self.assertEqual(hash3, md.refhash)

        md.filter_min_length(50000000000)
        self.assertEqual(md.nseqs, 0)
        self.assertFalse(np.any(md.mask))

    def test_minlength(self):
        with self.assertRaises(ValueError):
            Composition.from_file(self.io, minlength=3)

    def test_properties(self):
        composition = Composition.from_file(self.io, minlength=420)
        passed = list(filter(lambda x: len(x.sequence) >= 420, self.records))

        self.assertEqual(composition.nseqs, len(composition.metadata.identifiers))
        self.assertEqual(composition.nseqs, len(composition.metadata.lengths))

        self.assertTrue(composition.matrix.dtype, np.float32)
        self.assertEqual(composition.matrix.shape, (len(passed), 103))

        # Names
        self.assertEqual(
            list(composition.metadata.identifiers), [i.header for i in passed]
        )

        # Lengths
        self.assertTrue(np.issubdtype(composition.metadata.lengths.dtype, np.integer))
        self.assertEqual(
            [len(i.sequence) for i in passed], list(composition.metadata.lengths)
        )

    def test_save_load(self):
        buf = io.BytesIO()
        composition_1 = Composition.from_file(self.io)
        md1 = composition_1.metadata
        composition_1.save(buf)
        buf.seek(0)
        composition_2 = Composition.load(buf)
        md2 = composition_2.metadata

        self.assertTrue(np.all(composition_1.matrix == composition_2.matrix))
        self.assertTrue(np.all(md1.identifiers == md2.identifiers))
        self.assertTrue(np.all(md1.lengths == md2.lengths))
        self.assertTrue(np.all(md1.refhash == md2.refhash))
        self.assertTrue(np.all(md1.minlength == md2.minlength))

    def test_windows_newlines(self):
        rng = random.Random()
        buf1 = io.BytesIO()
        buf2 = io.BytesIO()
        for i in range(10):
            record = testtools.make_randseq(rng, 10, 20)
            buf1.write(b">" + record.header.encode())
            buf2.write(b">" + record.header.encode())
            buf1.write(b"\r\n")
            buf2.write(b"\n")
            buf1.write(record.sequence)
            buf2.write(record.sequence)
            buf1.write(b"\r\n")
            buf2.write(b"\n")

        buf1.seek(0)
        buf2.seek(0)
        comp1 = Composition.from_file(buf1)
        comp2 = Composition.from_file(buf2)

        self.assertEqual(comp1.metadata.refhash, comp2.metadata.refhash)
        self.assertTrue(np.all(comp1.matrix == comp2.matrix))



from typing import Optional, IO, Union
from pathlib import Path
import vamb.vambtools as _vambtools
from torch.utils.data.dataset import TensorDataset as _TensorDataset
from torch.utils.data import DataLoader as _DataLoader
from torch.nn.functional import softmax as _softmax
from torch import Tensor
import dadaptation
from torch import nn as _nn
from math import log as _log
from loguru import logger

__doc__ = """Encode a depths matrix and a tnf matrix to latent representation.

Creates a variational autoencoder in PyTorch and tries to represent the depths
and tnf in the latent space under gaussian noise.

Usage:
>>> vae = VAE(nsamples=6)
>>> dataloader = make_dataloader(depths, tnf, lengths)
>>> vae.trainmodel(dataloader)
>>> latent = vae.encode(dataloader) # Encode to latent representation
>>> latent.shape
(183882, 32)
"""

__cmd_doc__ = """Encode depths and TNF using a VAE to latent representation"""

import numpy as _np
import torch as _torch


def set_batchsize(
    data_loader: _DataLoader, batch_size: int, encode=False
) -> _DataLoader:
    """Effectively copy the data loader, but with a different batch size.

    The `encode` option is used to copy the dataloader to use before encoding.
    This will not drop the last minibatch whose size may not be the requested
    batch size, and will also not shuffle the data.
    """
    return _DataLoader(
        dataset=data_loader.dataset,
        batch_size=batch_size,
        shuffle=not encode,
        drop_last=False,
        num_workers=1 if encode else data_loader.num_workers,
        pin_memory=data_loader.pin_memory,
    )


def make_dataloader(
    abundance: _np.ndarray,
    tnf: _np.ndarray,
    lengths: _np.ndarray,
    batchsize: int = 256,
    destroy: bool = False,
    cuda: bool = False,
) -> _DataLoader:
    """Create a DataLoader from abundance, TNF and lengths.

    The dataloader is an object feeding minibatches of contigs to the VAE.
    The data are normalized versions of the input datasets.

    Inputs:
        abundance: Abundance matrix (N_contigs x N_samples)
        tnf: TNF matrix (N_contigs x N_TNF)
        lengths: Numpy array of sequence length (N_contigs)
        batchsize: Starting size of minibatches for dataloader
        destroy: Mutate abundance and tnf array in-place instead of making a copy.
        cuda: Pagelock memory of dataloader (use when using GPU acceleration)

    Outputs:
        DataLoader: An object feeding data to the VAE
    """

    if not isinstance(abundance, _np.ndarray) or not isinstance(tnf, _np.ndarray):
        raise ValueError("TNF and abundance must be Numpy arrays")

    if batchsize < 1:
        raise ValueError(f"Batch size must be minimum 1, not {batchsize}")

    if len(abundance) != len(tnf) or len(tnf) != len(lengths):
        raise ValueError(
            "Lengths of abundance, TNF and lengths arrays must be the same"
        )

    if not (abundance.dtype == tnf.dtype == _np.float32):
        raise ValueError("TNF and abundance must be Numpy arrays of dtype float32")

    # Copy if not destroy - this way we can have all following operations in-place
    # for simplicity
    if not destroy:
        abundance = abundance.copy()
        tnf = tnf.copy()

    # Normalize samples to have same depth
    sample_depths_sum = abundance.sum(axis=0)
    if _np.any(sample_depths_sum == 0):
        raise ValueError(
            "One or more samples have zero depth in all sequences, so cannot be depth normalized"
        )
    abundance *= 1_000_000 / sample_depths_sum
    total_abundance = abundance.sum(axis=1)

    # Normalize abundance to sum to 1
    n_samples = abundance.shape[1]
    zero_total_abundance = total_abundance == 0
    abundance[zero_total_abundance] = 1 / n_samples
    nonzero_total_abundance = total_abundance.copy()
    nonzero_total_abundance[zero_total_abundance] = 1.0
    abundance /= nonzero_total_abundance.reshape((-1, 1))

    # Normalize TNF and total abundance to make SSE loss work better
    total_abundance = _np.log(total_abundance.clip(min=0.001))
    _vambtools.zscore(total_abundance, inplace=True)
    _vambtools.zscore(tnf, axis=0, inplace=True)
    total_abundance.shape = (len(total_abundance), 1)

    # Create weights
    lengths = (lengths).astype(_np.float32)
    weights = _np.log(lengths).astype(_np.float32) - 5.0
    weights[weights < 2.0] = 2.0
    weights *= len(weights) / weights.sum()
    weights.shape = (len(weights), 1)

    ### Create final tensors and dataloader ###
    depthstensor = _torch.from_numpy(abundance)  # this is a no-copy operation
    tnftensor = _torch.from_numpy(tnf)
    total_abundance_tensor = _torch.from_numpy(total_abundance)
    weightstensor = _torch.from_numpy(weights)
    n_workers = 4 if cuda else 1
    dataset = _TensorDataset(
        depthstensor, tnftensor, total_abundance_tensor, weightstensor
    )
    dataloader = _DataLoader(
        dataset=dataset,
        batch_size=batchsize,
        drop_last=False,
        shuffle=True,
        num_workers=n_workers,
        pin_memory=cuda,
    )

    return dataloader


class VAE(_nn.Module):
    """Variational autoencoder, subclass of torch.nn.Module.

    Instantiate with:
        nsamples: Number of samples in abundance matrix
        nhiddens: list of n_neurons in the hidden layers [None=Auto]
        nlatent: Number of neurons in the latent layer [32]
        alpha: Approximate starting TNF/(CE+TNF) ratio in loss. [None = Auto]
        beta: Multiply KLD by the inverse of this value [200]
        dropout: Probability of dropout on forward pass [0.2]
        cuda: Use CUDA (GPU accelerated training) [False]

    vae.trainmodel(dataloader, nepochs batchsteps, modelfile)
        Trains the model, returning None

    vae.encode(self, data_loader):
        Encodes the data in the data loader and returns the encoded matrix.

    If alpha or dropout is None and there is only one sample, they are set to
    0.99 and 0.0, respectively
    """

    def __init__(
        self,
        nsamples: int,
        nhiddens: Optional[list[int]] = None,
        nlatent: int = 32,
        alpha: Optional[float] = None,
        beta: float = 200.0,
        dropout: Optional[float] = 0.2,
        cuda: bool = False,
        seed: int = 0,
    ):
        if nlatent < 1:
            raise ValueError(f"Minimum 1 latent neuron, not {nlatent}")

        if nsamples < 1:
            raise ValueError(f"nsamples must be > 0, not {nsamples}")

        # If only 1 sample, we weigh alpha and nhiddens differently
        if alpha is None:
            alpha = 0.15 if nsamples > 1 else 0.50

        if nhiddens is None:
            nhiddens = [512, 512] if nsamples > 1 else [256, 256]

        if dropout is None:
            dropout = 0.2 if nsamples > 1 else 0.0

        if any(i < 1 for i in nhiddens):
            raise ValueError(f"Minimum 1 neuron per layer, not {min(nhiddens)}")

        if beta <= 0:
            raise ValueError(f"beta must be > 0, not {beta}")

        if not (0 < alpha < 1):
            raise ValueError(f"alpha must be 0 < alpha < 1, not {alpha}")

        if not (0 <= dropout < 1):
            raise ValueError(f"dropout must be 0 <= dropout < 1, not {dropout}")

        _torch.manual_seed(seed)
        self.rng = _torch.Generator()
        self.rng.manual_seed(seed)
        super(VAE, self).__init__()

        # Initialize simple attributes
        self.usecuda = cuda
        self.nsamples = nsamples
        self.ntnf = 103
        self.alpha = alpha
        self.beta = beta
        self.nhiddens = nhiddens
        self.nlatent = nlatent
        self.dropout = dropout

        # Initialize lists for holding hidden layers
        self.encoderlayers = _nn.ModuleList()
        self.encodernorms = _nn.ModuleList()
        self.decoderlayers = _nn.ModuleList()
        self.decodernorms = _nn.ModuleList()

        # Add all other hidden layers
        for nin, nout in zip(
            # + 1 for the total abundance
            [self.nsamples + self.ntnf + 1] + self.nhiddens,
            self.nhiddens,
        ):
            self.encoderlayers.append(_nn.Linear(nin, nout))
            self.encodernorms.append(_nn.BatchNorm1d(nout))

        # Latent layers
        self.mu = _nn.Linear(self.nhiddens[-1], self.nlatent)

        # Add first decoding layer
        for nin, nout in zip([self.nlatent] + self.nhiddens[::-1], self.nhiddens[::-1]):
            self.decoderlayers.append(_nn.Linear(nin, nout))
            self.decodernorms.append(_nn.BatchNorm1d(nout))

        # Reconstruction (output) layer. + 1 for the total abundance
        self.outputlayer = _nn.Linear(self.nhiddens[0], self.nsamples + self.ntnf + 1)

        # Activation functions
        self.relu = _nn.LeakyReLU()
        self.softplus = _nn.Softplus()
        self.dropoutlayer = _nn.Dropout(p=self.dropout)

        if cuda:
            self.cuda()

    def _encode(self, tensor: Tensor) -> Tensor:
        tensors = list()

        # Hidden layers
        for encoderlayer, encodernorm in zip(self.encoderlayers, self.encodernorms):
            tensor = encodernorm(self.dropoutlayer(self.relu(encoderlayer(tensor))))
            tensors.append(tensor)

        # Latent layers
        mu = self.mu(tensor)

        # Note: We ought to also compute logsigma here, but we had a bug in the original
        # implementation of Vamb where logsigma was fixed to zero, so we just remove it.

        return mu

    # sample with gaussian noise
    def reparameterize(self, mu: Tensor) -> Tensor:
        epsilon = _torch.randn(mu.size(0), mu.size(1))

        if self.usecuda:
            epsilon = epsilon.cuda()

        epsilon.requires_grad = True

        latent = mu + epsilon

        return latent

    def _decode(self, tensor: Tensor) -> tuple[Tensor, Tensor, Tensor]:
        tensors = list()

        for decoderlayer, decodernorm in zip(self.decoderlayers, self.decodernorms):
            tensor = decodernorm(self.dropoutlayer(self.relu(decoderlayer(tensor))))
            tensors.append(tensor)

        reconstruction = self.outputlayer(tensor)

        # Decompose reconstruction to depths and tnf signal
        depths_out = reconstruction.narrow(1, 0, self.nsamples)
        tnf_out = reconstruction.narrow(1, self.nsamples, self.ntnf)
        abundance_out = reconstruction.narrow(1, self.nsamples + self.ntnf, 1)

        depths_out = _softmax(depths_out, dim=1)

        return depths_out, tnf_out, abundance_out

    def forward(
        self, depths: Tensor, tnf: Tensor, abundance: Tensor
    ) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        tensor = _torch.cat((depths, tnf, abundance), 1)
        mu = self._encode(tensor)
        latent = self.reparameterize(mu)
        depths_out, tnf_out, abundance_out = self._decode(latent)

        return depths_out, tnf_out, abundance_out, mu

    def calc_loss(
        self,
        depths_in: Tensor,
        depths_out: Tensor,
        tnf_in: Tensor,
        tnf_out: Tensor,
        abundance_in: Tensor,
        abundance_out: Tensor,
        mu: Tensor,
        weights: Tensor,
    ) -> tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
        ab_sse = (abundance_out - abundance_in).pow(2).sum(dim=1)
        # Add 1e-9 to depths_out to avoid numerical instability.
        ce = -((depths_out + 1e-9).log() * depths_in).sum(dim=1)
        sse = (tnf_out - tnf_in).pow(2).sum(dim=1)
        kld = 0.5 * (mu.pow(2)).sum(dim=1)

        # Avoid having the denominator be zero
        if self.nsamples == 1:
            ce_weight = 0.0
        else:
            ce_weight = ((1 - self.alpha) * (self.nsamples - 1)) / (
                self.nsamples * _log(self.nsamples)
            )

        ab_sse_weight = (1 - self.alpha) * (1 / self.nsamples)
        sse_weight = self.alpha / self.ntnf
        kld_weight = 1 / (self.nlatent * self.beta)
        weighed_ab = ab_sse * ab_sse_weight
        weighed_ce = ce * ce_weight
        weighed_sse = sse * sse_weight
        weighed_kld = kld * kld_weight
        reconstruction_loss = weighed_ce + weighed_ab + weighed_sse
        loss = (reconstruction_loss + weighed_kld) * weights

        return (
            loss.mean(),
            weighed_ab.mean(),
            weighed_ce.mean(),
            weighed_sse.mean(),
            weighed_kld.mean(),
        )

    def trainepoch(
        self,
        data_loader: _DataLoader,
        epoch: int,
        optimizer,
        batchsteps: list[int],
    ) -> _DataLoader[tuple[Tensor, Tensor, Tensor]]:
        if len(data_loader.dataset.tensors[0]) < 2:
            raise ValueError(
                "Cannot train on a dataset with fewer than 2 sequences, but got "
                f"{len(data_loader.dataset.tensors[0])} sequences. "
                "If you are trying to fit a DL model to this few sequences, "
                "something probably went wrong in your pipeline."
            )

        self.train()

        epoch_loss = 0.0
        epoch_kldloss = 0.0
        epoch_sseloss = 0.0
        epoch_celoss = 0.0
        epoch_absseloss = 0.0

        if epoch in batchsteps:
            data_loader = set_batchsize(
                data_loader,
                data_loader.batch_size * 2,  # type:ignore
            )

        for depths_in, tnf_in, abundance_in, weights in data_loader:
            depths_in.requires_grad = True
            tnf_in.requires_grad = True
            abundance_in.requires_grad = True

            if self.usecuda:
                depths_in = depths_in.cuda()
                tnf_in = tnf_in.cuda()
                abundance_in = abundance_in.cuda()
                weights = weights.cuda()

            optimizer.zero_grad()

            depths_out, tnf_out, abundance_out, mu = self(
                depths_in, tnf_in, abundance_in
            )

            loss, ab_sse, ce, sse, kld = self.calc_loss(
                depths_in,
                depths_out,
                tnf_in,
                tnf_out,
                abundance_in,
                abundance_out,
                mu,
                weights,
            )

            loss.backward()
            optimizer.step()

            epoch_loss += loss.data.item()
            epoch_kldloss += kld.data.item()
            epoch_sseloss += sse.data.item()
            epoch_celoss += ce.data.item()
            epoch_absseloss += ab_sse.data.item()

        logger.info(
            "\t\tEpoch: {:>3}  Loss: {:.5e}  CE: {:.5e}  AB: {:.5e}  SSE: {:.5e}  KLD: {:.5e}  Batchsize: {:>4}".format(
                epoch + 1,
                epoch_loss / len(data_loader),
                epoch_celoss / len(data_loader),
                epoch_absseloss / len(data_loader),
                epoch_sseloss / len(data_loader),
                epoch_kldloss / len(data_loader),
                data_loader.batch_size,
            )
        )

        self.eval()
        return data_loader

    def encode(self, data_loader) -> _np.ndarray:
        """Encode a data loader to a latent representation with VAE

        Input: data_loader: As generated by train_vae

        Output: A (n_contigs x n_latent) Numpy array of latent repr.
        """

        self.eval()

        new_data_loader = set_batchsize(
            data_loader, data_loader.batch_size, encode=True
        )

        depths_array, _, _, _ = data_loader.dataset.tensors
        length = len(depths_array)

        # We make a Numpy array instead of a Torch array because, if we create
        # a Torch array, then convert it to Numpy, Numpy will believe it doesn't
        # own the memory block, and array resizes will not be permitted.
        latent = _np.empty((length, self.nlatent), dtype=_np.float32)

        row = 0
        with _torch.no_grad():
            for depths, tnf, ab, _ in new_data_loader:
                # Move input to GPU if requested
                if self.usecuda:
                    depths = depths.cuda()
                    tnf = tnf.cuda()
                    ab = ab.cuda()

                # Evaluate
                _, _, _, mu = self(depths, tnf, ab)

                if self.usecuda:
                    mu = mu.cpu()

                latent[row : row + len(mu)] = mu
                row += len(mu)

        assert row == length
        _vambtools.mask_lower_bits(latent, 12)
        return latent

    def save(self, filehandle):
        """Saves the VAE to a path or binary opened file. Load with VAE.load

        Input: Path or binary opened filehandle
        Output: None
        """
        state = {
            "nsamples": self.nsamples,
            "alpha": self.alpha,
            "beta": self.beta,
            "dropout": self.dropout,
            "nhiddens": self.nhiddens,
            "nlatent": self.nlatent,
            "state": self.state_dict(),
        }

        _torch.save(state, filehandle)

    @classmethod
    def load(
        cls, path: Union[IO[bytes], str], cuda: bool = False, evaluate: bool = True
    ):
        """Instantiates a VAE from a model file.

        Inputs:
            path: Path to model file as created by functions VAE.save or
                  VAE.trainmodel.
            cuda: If network should work on GPU [False]
            evaluate: Return network in evaluation mode [True]

        Output: VAE with weights and parameters matching the saved network.
        """

        # Forcably load to CPU even if model was saves as GPU model
        dictionary = _torch.load(
            path, map_location=lambda storage, loc: storage, weights_only=True
        )

        nsamples = dictionary["nsamples"]
        alpha = dictionary["alpha"]
        beta = dictionary["beta"]
        dropout = dictionary["dropout"]
        nhiddens = dictionary["nhiddens"]
        nlatent = dictionary["nlatent"]
        state = dictionary["state"]

        vae = cls(nsamples, nhiddens, nlatent, alpha, beta, dropout, cuda)
        vae.load_state_dict(state)

        if cuda:
            vae.cuda()

        if evaluate:
            vae.eval()

        return vae

    def trainmodel(
        self,
        dataloader: _DataLoader[tuple[Tensor, Tensor, Tensor]],
        nepochs: int = 500,
        batchsteps: Optional[list[int]] = [25, 75, 150, 300],
        modelfile: Union[None, str, Path, IO[bytes]] = None,
    ):
        """Train the autoencoder from depths array and tnf array.

        Inputs:
            dataloader: DataLoader made by make_dataloader
            nepochs: Train for this many epochs before encoding [500]
            batchsteps: None or double batchsize at these epochs [25, 75, 150, 300]
            modelfile: Save models to this file if not None [None]

        Output: None
        """
        if nepochs < 1:
            raise ValueError("Minimum 1 epoch, not {nepochs}")

        if batchsteps is None:
            batchsteps_set: set[int] = set()
        else:
            # First collect to list in order to allow all element types, then check that
            # they are integers
            batchsteps = list(batchsteps)
            if not all(isinstance(i, int) for i in batchsteps):
                raise ValueError("All elements of batchsteps must be integers")
            if max(batchsteps, default=0) >= nepochs:
                raise ValueError("Max batchsteps must not equal or exceed nepochs")
            batchsteps_set = set(batchsteps)

        # Get number of features
        # Following line is un-inferrable due to typing problems with DataLoader
        ncontigs, nsamples = dataloader.dataset.tensors[0].shape  # type: ignore
        optimizer = dadaptation.DAdaptAdam(self.parameters(), decouple=True)

        logger.info("\tNetwork properties:")
        logger.info(f"\t    CUDA: {self.usecuda}")
        logger.info(f"\t    Alpha: {self.alpha}")
        logger.info(f"\t    Beta: {self.beta}")
        logger.info(f"\t    Dropout: {self.dropout}")
        logger.info(f"\t    N hidden: {', '.join(map(str, self.nhiddens))}")
        logger.info(f"\t    N latent: {self.nlatent}")
        logger.info("\tTraining properties:")
        logger.info(f"\t    N epochs: {nepochs}")
        logger.info(f"\t    Starting batch size: {dataloader.batch_size}")
        batchsteps_string = (
            ", ".join(map(str, sorted(batchsteps_set))) if batchsteps_set else "None"
        )
        logger.info(f"\t    Batchsteps: {batchsteps_string}")
        logger.info(f"\t    N sequences: {ncontigs}")
        logger.info(f"\t    N samples: {nsamples}")

        # Train
        for epoch in range(nepochs):
            dataloader = self.trainepoch(
                dataloader, epoch, optimizer, sorted(batchsteps_set)
            )

        # Save weights - Lord forgive me, for I have sinned when catching all exceptions
        if modelfile is not None:
            try:
                self.save(modelfile)
            except:
                pass

        return None



__doc__ = """Create kernel for use in kmer frequencies.
Method copied from https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2765972/

Principle:
There are 256 tetranucleotides, so a frequency distribution (tetranucleotide frequency, TNF)
is a length 256 vector. But the individual TNFs are not independent. For example, AAAT
must correlate highly with AATA. The TNFs are subject to at least 3 linear constrains:

1) The vector must sum to one. We simply shift the TNF down by 1/256 to make it sum to zero
for simplicity instead.
2) We cannot distinguish between a kmer and its reverse complement because the sequencede
strand is arbitrary. So we must count e.g. AGAT as one half of AGAT and one half ATCT.
So each kmer's frequency is the same as its reverse-complement.
3) Every time a kmer is observed, the next kmer must have three overlapping nucleotides.
E.g. every observation of AGAT is followed by GATA, GATC, GATG or GATT. Same for previous
kmer. in other words, sum(xABC) = sum(ABCx).
This is not true right at the ends of the sequences because the kmers stop eventually, but
that can be considered a measurement error, and we don't care about it.

We list these linear constrains and produce kernel L that works on tnf matrix T such that
TL = P, a smaller projected TNF space.

Notably, for constraint 2 to be true, we need to average the frequency between a kmer
and its reverse complement. We can do this with a matrix multiply with an averaging kernel
R. So:

P = (TR)L = T(RL) = TK

We thus calculate K = RL and save this for use in Vamb for projection.
"""

from os.path import abspath, dirname, join
import numpy as np
import itertools
from scipy.linalg import null_space


def reverse_complement(nuc):
    table = str.maketrans("ACGT", "TGCA")
    return nuc[::-1].translate(table)


def all_kmers(k):
    for i in itertools.product("ACGT", repeat=k):
        yield ("".join(i))


def create_projection_kernel():
    indexof = {kmer: i for i, kmer in enumerate(all_kmers(4))}
    linear_equations = list()

    # Constraint one: Frequencies sum to one (or in this scaled case, zero)
    linear_equations.append([1] * 256)

    # Constaint two: Frequencies are same as that of reverse complement
    for kmer in all_kmers(4):
        revcomp = reverse_complement(kmer)

        # Only look at canonical kmers - this makes no difference
        if kmer >= revcomp:
            continue

        line = [0] * 256
        line[indexof[kmer]] = 1
        line[indexof[revcomp]] = -1
        linear_equations.append(line)

    # Constraint three: sum(ABCx) = sum(xABC)
    for trimer in all_kmers(3):
        line = [0] * 256
        for suffix in "ACGT":
            line[indexof[trimer + suffix]] += 1
        for prefix in "ACGT":
            line[indexof[prefix + trimer]] += -1
        linear_equations.append(line)

    linear_equations = np.array(linear_equations)
    kernel = null_space(linear_equations).astype(np.float32)
    assert kernel.shape == (256, 103)
    return kernel


def create_rc_kernel():
    indexof = {kmer: i for i, kmer in enumerate(all_kmers(4))}
    rc_matrix = np.zeros((256, 256), dtype=np.float32)
    for col, kmer in enumerate(all_kmers(4)):
        revcomp = reverse_complement(kmer)
        rc_matrix[indexof[kmer], col] += 0.5
        rc_matrix[indexof[revcomp], col] += 0.5

    return rc_matrix


def create_dual_kernel():
    return np.dot(create_rc_kernel(), create_projection_kernel())


dual_kernel = create_dual_kernel()

# Prevent overwriting kernel when running tests
if __name__ == "__main__":
    path = join(dirname(dirname(abspath(__file__))), "vamb", "kernel.npz")
    np.savez_compressed(path, dual_kernel)



"""Vamb - Variational Autoencoders for Metagenomic Binning
Documentation: https://github.com/RasmussenLab/vamb/
"""

# TODO: Pyhmmer is compiled with -funsafe-math-optimizations, which toggles some
# flag in the CPU controlling float subnormal behaviour.
# This causes a warning in NumPy.
# This is not an issue in Vamb (I think), so we silence the warning here as a
# temporary fix.
# See https://github.com/althonos/pyhmmer/issues/71
import warnings

warnings.filterwarnings("ignore", category=UserWarning, module="numpy")

from . import vambtools
from . import parsebam
from . import parsecontigs
from . import parsemarkers
from . import taxonomy
from . import cluster
from . import encode
from . import aamb_encode
from . import semisupervised_encode
from . import hloss_misc
from . import taxvamb_encode
from . import reclustering

from importlib.metadata import version as get_version
from loguru import logger

__version_str__ = get_version("vamb")
logger.remove()

__all__ = [
    "vambtools",
    "parsebam",
    "parsecontigs",
    "parsemarkers",
    "taxonomy",
    "cluster",
    "encode",
    "aamb_encode",
    "semisupervised_encode",
    "taxvamb_encode",
    "hloss_misc",
    "reclustering",
]



import numpy as np
import argparse
from vamb.vambtools import RefHasher
from pathlib import Path


def abundances_mask(headers: Path, mask_refhash: Path, min_contig_size: int):
    """# Using the headers above, compute the mask and the refhash"""

    mask = []
    identifiers = []

    with open(headers) as file:
        for line in file:
            # SN:S27C112075   LN:2239
            (sn, ln) = line.split("\t")
            if sn[:3] != "SN:" or ln[:3] != "LN:":
                raise ValueError("Unknown format")
            passed = int(ln[3:]) >= min_contig_size
            mask.append(passed)
            if passed:
                identifiers.append(sn[3:])

    np.savez_compressed(
        mask_refhash,
        mask=np.array(mask, dtype=bool),
        refhash=RefHasher.hash_refnames(identifiers),
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--h", type=Path, help=" Headers file")
    parser.add_argument("--msk", type=Path, help="mask refhash")

    parser.add_argument("--minsize", type=int, help="min contig size")

    opt = parser.parse_args()

    abundances_mask(opt.h, opt.msk, opt.minsize)



import unittest

# For CAMI dataset, compute comp, abundance, taxonomy, markers
# Subset to e.g. 5 genera plus a few unclassified contigs

# FASTA
# Comp
# Abundance
# Markers
# Taxonomy
# Latent
# Refined taxonomy


class TestKmeansReclustering(unittest.TestCase):
    pass
    # Make markers + lengths
    # Make taxonomy
    # Create latent

    # Initial clustering


class TestDBScanReclustering(unittest.TestCase):
    # It produces disjoint clusters, a subset of the input points
    pass
